/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This is the main loose module and master emp dispatcher
 * Messages arrive here and are dispatched to child processes (slaves) which
 * control thoes nodes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <libloose.h>
#include <emp_module_harness.h>
#include <config.h>
#include <prototypes.h>
#include <signal.h>

/* This number determines the number of nodes that a slave process may control*/
int	num_nodes_per_slave = 512;

static void     signal_handler(int I_signo);
static void			config_cb(__uint32_t job_id, int client_fd, char *msg);
static void			deinit_cb(__uint32_t job_id, int client_fd, char *msg);
static void			discon_cb(__uint32_t job_id, int client_fd, char *msg);
static void			ipc_cb(char				*node_id,
											 __uint32_t	job_id,
											 int				client_fd,
											 char				*msg);
static NODELIST	*node_create(char *node_id);
static NODELIST	*lookup_node(char *node_id);

extern int errno;

NODELIST	*nodelist_head = NULL;
NODELIST	*nodelist_tail = NULL;

int main(argc, argv)
int		argc;
char	**argv;
{
	KNOWN_UNUSED_PARAM(argc);
	KNOWN_UNUSED_PARAM(argv);

  signal (SIGSEGV, signal_handler);

	lm_init();
	lm_nexxus_connect();
	if (lm_register("EMP",
									"Emergency Management Port Module",
									"Module to interface with a remote Emergency Management Port",
									"San Mehat (nettwerk@valinux.com)",
									"EMP",
									2,
									0) < 0)
		{
		printf("Unable to register module with Nexxus\n");
		exit(-1);
		}

	lm_main_loop(config_cb, deinit_cb, discon_cb, ipc_cb);
	exit(0);
}

static void config_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;
	char			*cmd_module;
	char			*node_id;
	char			logmsg[255];

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	cmd_module = strtok(msg, ":");
	node_id = strtok(NULL, ":");

	node = lookup_node(node_id);

	if (!strcasecmp(cmd_module, "NODE"))
		{
		node = node_create(node_id);
		slave_assign_and_notify(node);
		}
	else if (!strcasecmp(cmd_module, "END_NODE"))
		{
		}
	else if (!strcasecmp(cmd_module, "ADDITION"))
		{
		node = node_create(node_id);
		slave_assign_and_notify(node);
		return;
		}
	else if (!strcasecmp(cmd_module, "DELETION"))
		{
		slave_notify_del_node(node);
		node_destroy(node);
		return;
		}
	else if (!strcasecmp(cmd_module, "RENAME"))
		{
		char	*new_name;
		
		if (!(new_name = strtok(NULL,":")))
			{
			lm_log("Malformatted cfg msg\n");
			return;
			}
		slave_notify_rename_node(node, new_name);
		strcpy(node->node_id, new_name);
		return;
		}
	else if (!strcasecmp(cmd_module, "GLOBAL"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			lm_log("Malformatted cfg msg\n");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			lm_log("Malformatted cfg msg\n");
			return;
			}
		/* Any global variables we care about here... ie: */
		if (!strcasecmp(variable, "IP_ADDRESS"))
			{
			slave_notify_variable(node, variable, value);
			return;
			}
		}
	else if (!strcasecmp(cmd_module, "EMP"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			lm_log("Malformatted cfg msg");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			lm_log("Malformatted cfg msg\n");
			return;
			}
		/* Process any local variables here */
		if (!strcasecmp(variable, "SERIAL_PORT"))
			{
			slave_notify_variable(node, variable, value);
			return;
			}
		else if (!strcasecmp(variable, "PASSWORD"))
			{
			slave_notify_variable(node, variable, value);
			return;
			}
		else if (!strcasecmp(variable, "SERIAL_NUMBER"))
			{
			return; /* Toss the serial number.. backwards compatability */
			}
		sprintf(logmsg,
						"Received unknown local module variable '%s'",
						variable);
		lm_log(logmsg);
		}
	else
		{
		sprintf(logmsg,"Received unknown config event <%s>",cmd_module);
		lm_log(logmsg);
		}
}

static void deinit_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);

	slave_deinit();
	lm_nexxus_disconnect();
	exit(0);
}

static void discon_cb(__uint32_t job_id, int client_fd, char *msg)
{
	slave_notify_disconnection(client_fd);
	KNOWN_UNUSED_PARAM(job_id);
  KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);
}

static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg)
{
	char			*cmd, *p, *toss;
	char			response_msg[255], saved_msg[255];
	char			slave_msg[255];
	NODELIST	*node;

	strncpy(saved_msg,msg,(sizeof(saved_msg)-1));

	if (!(cmd = strtok(msg, ":")))
		goto out_malformatted;

	if (!(toss = strtok(NULL, ":")))
		goto out_malformatted;

	if (!(node = lookup_node(node_id)))
		{
		sprintf(response_msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_FOUND",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		return;
		}
	if (!node->slave)
		{
		sprintf(response_msg, "Slave not assigned for node %s",node->node_id);
		lm_log(response_msg);
		sprintf(response_msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_READY",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		return;
		}

	p = &saved_msg[((toss - msg) + strlen(toss)+1)];
	sprintf(slave_msg,
					"NODE_IPC:%s:%d:%d:%s:%s",
					node->node_id,
					job_id,
					client_fd,
					cmd,
					p);
	slave_send(node->slave, slave_msg);
	return;

out_malformatted:
	sprintf(response_msg,
					"%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
					job_id,
					client_fd);
	lm_send_to_nexxus(response_msg);
	return;
}

static NODELIST	*node_create(char *node_id)
{
	NODELIST	*new;

	if (!(new = (NODELIST	*) malloc(sizeof(NODELIST))))
		return(0);
	memset(new, 0, sizeof(NODELIST));
	if (!nodelist_head)
		nodelist_head = nodelist_tail = new;
	else
		{
		nodelist_tail->next = new;
		new->prev = nodelist_tail;
		nodelist_tail = new;
		}
	strcpy(new->node_id, node_id);
	return(new);
}

static NODELIST	*lookup_node(char *node_id)
{
	NODELIST	*scan;

	if (!node_id)
		return(NULL);
	scan = nodelist_head;
	while(scan)
		{
		if (!strcmp(node_id, scan->node_id))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

void node_destroy(NODELIST	*node)
{
  if (node == nodelist_head)
    {
    nodelist_head = nodelist_head->next;
    if (nodelist_head)
      nodelist_head->prev = NULL;
    else
      nodelist_tail = NULL;
    } 
  else if (node == nodelist_tail)
    {
    nodelist_tail = nodelist_tail->prev;
    if (nodelist_tail)
      nodelist_tail->next = NULL;
    else
      nodelist_head = NULL;
    } 
  else
    {
    node->next->prev = node->prev;
    node->prev->next = node->next;
    }
  free(node);
}

static void signal_handler(int I_signo)
{
  extern CONTEXT          *context_head;
  CONTEXT *ctx = context_head;
	char			logmsg[255];
	
  if (I_signo==SIGSEGV)
    {
		while(ctx)
			{
			if (ctx->thread == pthread_self())
				{	
				sprintf(logmsg,
						"Slave thread for node %s has crashed",
						ctx->node_id);
				lm_log(logmsg);
				sleep(5);
				abort();
				}
			else if (ctx->driver_thread == pthread_self())
				{
				sprintf(logmsg,
						"Driver thread for node %s has crashed",
						ctx->node_id);
				lm_log(logmsg);
				sleep(5);
				abort();
				}
			ctx = ctx->next;
		  }
    }
}
