/*
 * Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:      Dean Johnson (dtj@sgi.com)
 *
 * Purpose:     
 *
 * This Module is based heavily on the other loose modules by San Mehat
 * <nettwerk@valinux.com> and Zac Sprackett <zsprackett@valinux.com>.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <netdb.h>
#include <libloose.h>
#include <msc_slave_context.h>
#include <msc_module_harness.h>
#include <config.h>
#include <pthread.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(SLAVE_CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_msc_ipc(SLAVE_CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static void	set_addr(SLAVE_CONTEXT *ctx, char *addr);
static void	set_device(SLAVE_CONTEXT *ctx, char *addr);
static FILE	*do_rsh(char *addr, char *user, char *cmd);

void	*thread_start(void *);

#ifdef	notdef
typedef struct {
	FILE    *inputfd;
	int     replyfd;
	unsigned int    jobid;
} DeferredInput;
#endif	/*notdef*/

static void
send_log(int fd,char *msg)
{
	char	buf[1024];

	sprintf(buf,"0:0:LOG:%s",msg);
	send_msg(fd, buf);
}

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern SLAVE_CONTEXT	*context_head;

static void
thread_cleanup(void *data)
{
	SLAVE_CONTEXT *ctx = (SLAVE_CONTEXT *) data;
	
#ifdef THREAD_DEBUG
	printf("Thread %s cleaning up\n",ctx->node_id);
#endif
#ifdef THREAD_DEBUG
	printf("Thread %s clean\n",ctx->node_id);
#endif
	ctx->refcount--;
}

void *
thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	SLAVE_CONTEXT	*ctx = (SLAVE_CONTEXT *) data;
	struct timeval	to;
	fd_set	read_fds;
	int	max, rc;
	char	log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,"Thread %s starting (pid %d)",ctx->node_id,getpid());
	send_log(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while (1) {
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0) {
			sprintf(log_msg,"Thread %s select() failed (%s)",
				ctx->node_id,strerror(errno));
			send_log(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
		}
		if (!rc) {
			pthread_testcancel();
		}
		else if (FD_ISSET(ctx->thread_ipc_read, &read_fds)) {
			process_msg_from_slave(ctx);
		}
		pthread_testcancel();
	}
}

static void
process_msg_from_slave(SLAVE_CONTEXT *ctx)
{
	char	*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char	*ipc_cmd, *ipc_args, *variable, *value;
	char	*new_name;
	int	len, rc, client_fd;
	__uint32_t	job_id;
	char	*last = (char *) 0x00;
	char	log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0) {
		sprintf(log_msg,"Thread %s unable to read msg from slave (%s)",
			ctx->node_id,strerror(-rc));
		send_log(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
	}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",ctx->node_id,msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC")) {
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_msc_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
	}
	else if (!strcasecmp(cmd, "NODE_VAR")) {
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;
		if (!(value = strtok_r(NULL, ":", &last)))
			goto out;
		if (!strcasecmp(variable,"IP_ADDRESS"))
			set_addr(ctx, value);
		if (!strcasecmp(variable,"SERIAL_PORT"))
			set_device(ctx, value);
	}
	else if (!strcasecmp(cmd, "CLIENT_DIS")) {
		char	*fd;
		int	client_fd;

		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
	}
	else if (!strcasecmp(cmd, "NODE_RENAME")) {
		char	logfile_old[255];
		char	logfile_new[255];

		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
		/* Rename the logfile */
		sprintf(logfile_old,"%s/vacm_msc_alert_log.%s",LOGDIR,
			ctx->node_id);
		free(ctx->node_id);
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		sprintf(logfile_new,"%s/vacm_msc_alert_log.%s",LOGDIR,
			ctx->node_id);
		if (rename(logfile_old, logfile_new) < 0) {
			sprintf(log_msg,"Thread %s unable to rename logfile (%m)",
				ctx->node_id);
			send_log(ctx->thread_ipc_write, log_msg);
		}
	}
	else	{
		sprintf(log_msg,"Thread %s received unsupported cmd from slave (%s)",
			ctx->node_id,cmd);
		send_log(ctx->thread_ipc_write, log_msg);
	}

out:
	free(saved_buffer);
	free(msg_buffer);
}

static void
set_addr(SLAVE_CONTEXT *ctx, char *addr)
{
	if (ctx->addr)
		free(ctx->addr);
	ctx->addr = malloc((strlen(addr)+1));
	strcpy(ctx->addr, addr);
}

static void
set_device(SLAVE_CONTEXT *ctx, char *addr)
{
	strcpy(ctx->device_address, addr);
	ctx->msc_fd = open(ctx->device_address,O_RDWR);
}

static int
getPingInfo(SLAVE_CONTEXT *ctx,FILE *inf)
{
	char	line[1024];
//	char	resp_msg[1024];
	int	lines = 0;
	char	*cp;
	KNOWN_UNUSED_PARAM(ctx);

	while (fgets(line,1023,inf) != NULL) {
		if (line[0] == '\n') continue;
		if ((cp = strstr(line,"packets received")) != NULL) {
			/*
			 *        backup to the comma and try
			 *        and find number of successful
			 *        packets.
			 */
			while (cp > line && *cp != ',') cp--;
			if (*cp == ',') {
				cp++;
				while (*cp && !isdigit(*cp)) cp++;
				return(*cp == '0' ? 0 : 1);
			}
			lines++;
			continue;
		}
		else        {
		}
		lines++;
	}
	return(1);
}

/*
 * Process an IPC message
 */
static void
process_msc_ipc(SLAVE_CONTEXT *ctx,__uint32_t job_id,int client_fd,char *ipc_cmd,
		char *ipc_args)
{
	FILE	*inf;
	char	resp_msg[255];
	char	cmd_msg[255];
	char	*last = (char *) 0x00;
	int	done = 0;

	if (!strcasecmp(ipc_cmd, "CONFIGURATION")) {
		char	*device_addr;
		char	*host;

		if (!(device_addr = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(host = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;

		strcpy(ctx->device_address,device_addr);
		strcpy(ctx->addr,host);

		sprintf(resp_msg,"0:0:SET:%s:MSC:SERIAL_PORT:%s:",ctx->node_id,
			ctx->device_address);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg,"0:0:SET:%s:MSC:IP_ADDRESS:%s:",ctx->node_id,
			ctx->addr);
		send_msg(ctx->thread_ipc_write,resp_msg);

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);	
		goto out_ok;
	}

	/*--------------------------------------------------------------*
	 *			REAL COMMANDS				*
	 *--------------------------------------------------------------*/

	/* Begin real commands */
/*#*/	if (!strcasecmp(ipc_cmd, "INQUIRY")) { // //Inquiry
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:%s:%s:",job_id,client_fd,
			ctx->node_id,ctx->addr);
		send_msg(ctx->thread_ipc_write, resp_msg);
	}
	else if (!strcasecmp(ipc_cmd, "NODE_INV_INFO")) { // Node Inventory
		char	msg_buffer[4096];
		FILE	*rem = NULL;
		char    cmd[1024];
                char *os;
                char *version;

//		if (ctx->agent != NULL && !strcasecmp(ctx->agent,"msc"))
			if (!(rem=do_rsh(ctx->addr,"root","uname -rs")))
				goto out_malformatted;

                fgets(msg_buffer,sizeof(msg_buffer),rem);
                os = strtok_r(msg_buffer," ",&last);
                version = strtok_r(NULL,"\n",&last);
		sprintf(cmd,"%d:%d:FOR:NODEINVINFO:%s:%s:",job_id,client_fd,os,
			version);
		send_msg(ctx->thread_ipc_write,cmd);
		fclose(rem);
	}
        else if (!strcasecmp(ipc_cmd,"CHASSIS_STATUS"))   {
//		if (ctx->agent != NULL && !strcasecmp(ctx->agent,"msc")) {
			sprintf(resp_msg,"Requesting chassis status for <%s>",
				ctx->addr);
			send_log(ctx->thread_ipc_write,resp_msg);
			sprintf(cmd_msg,"/bin/ping -c 1 %s",ctx->addr);
	                if ((inf = popen(cmd_msg,"r")) == NULL) {
				send_log(ctx->thread_ipc_write,
					 "ping popen failed!!!!!");
				fprintf(stderr,"couldn't popen()!\n");
//				return(NODE_NOT_FOUND);
				goto out_malformatted;
			}
			sprintf(resp_msg,"%d:%d:FOR:CSTATUS:%s:",job_id,
				client_fd,getPingInfo(ctx,inf) ? "YES" : "NO");
                        send_msg(ctx->thread_ipc_write,resp_msg);
			sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
			send_msg(ctx->thread_ipc_write, resp_msg);	
			fclose(inf);
			goto out_ok;
//		}
	}

        /*--------------------------------------------------------------*
         *                        CHASSIS CONTROL                       *
         *--------------------------------------------------------------*/

#define SEND_CMD(s)							\
		send_msg(ctx->msc_fd,"\024pas none\n");			\
		sleep(1);						\
		send_msg(ctx->msc_fd,s);				\
		done = 1

        else if (!strcasecmp(ipc_cmd,"POWER_OFF"))   {
		SEND_CMD("\024pwr d0\n"); 
	}
	else if (!strcasecmp(ipc_cmd,"POWER_ON"))    {
		SEND_CMD("\024pwr u\n"); 
	}
	else if (!strcasecmp(ipc_cmd,"POWER_CYCLE")) { 
		SEND_CMD("\024pwr c\n");  
	}
	else if (!strcasecmp(ipc_cmd,"HARD_RESET"))  { 
		SEND_CMD("\024rst\n");    
	}
	else	{
		sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,
			strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
	}
	sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
	send_msg(ctx->thread_ipc_write, resp_msg);	

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}


/*
 * Send a message. This function is threadsafe
 */
static int
send_msg(int fd, char *msg)
{
	__uint32_t	len;
	__uint32_t	net_len;
	int		rc;

	if (fd == -1)
		return(-EBADF);
	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    		return(rc);
  	if ((rc = write (fd, (void *) msg, len)) < 0)
    		return(rc);
  	return(0);
}

/*
 * Receive a message. This function is threadsafe
 */
static int
recv_msg(int fd, char **buffer, __uint32_t *len)
{
	__uint32_t	net_len;
	__uint32_t	brtr;
	char	*offset;
	int	rc;

	if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    		return(rc);
	else if (!rc)
		return(-1);
	*len = ntohl(net_len);
	if (!(*buffer = malloc(*len)))
    		return(-ENOMEM);
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while (brtr) {
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
	}
	return(0);
}

static FILE *
do_rsh(char *addr, char *user, char *cmd)
{
	char    tmp[64];
	char    *address = tmp;
	int     remote;

	strcpy(tmp,addr);

#define SHELL 514

	remote = rcmd(&address,SHELL,"root",user,cmd,0);
	if (remote<0)
		return(NULL);

	return(fdopen(remote,"r"));
}
