
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 *  with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <pthread.h>
#include <libloose.h>
#include <rsh_module_harness.h>
#include <rsh_slave_context.h>
#include <config.h>
#include <prototypes.h>

static int read_message(int, char **, __uint32_t *);
static void process_msg_from_master(int, int);
static void process_msg_from_thread(CONTEXT *, int, int);
static int send_message(int,char *);

extern CONTEXT *context_head;
extern CONTEXT *context_tail;

// #define SLAVE_DEBUG
void slave_machine_start(int read_fd, int write_fd)
{
	struct timeval	to;
	pid_t						pid;
	fd_set					read_fds;
	int							max, rc;
	CONTEXT					*ctx;

	pid = getpid();
#ifdef SLAVE_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Slave pid %d starting",
				 	pid);
	send_message(write_fd, log_msg);
#endif

	context_init();

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(read_fd, &read_fds);
		max = read_fd;

		ctx = context_head;
		while(ctx)
			{
			FD_SET(ctx->slave_ipc_read, &read_fds);
			if (max < ctx->slave_ipc_read)
				max = ctx->slave_ipc_read;
			ctx = ctx -> next;
			}
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			printf("Slave %d select failed\n",pid);
			continue;
			}
		else if (!rc)
			{
			continue;
			}
		if (FD_ISSET(read_fd, &read_fds))
			{
			process_msg_from_master(read_fd, write_fd);
			continue;
			}
		ctx = context_head;
		while(ctx)
			{
			if (FD_ISSET(ctx->slave_ipc_read, &read_fds))
				process_msg_from_thread(ctx, read_fd, write_fd);
			ctx = ctx -> next;
			}
		}
}

static void process_msg_from_master(int r_fd, int w_fd)
{
	char				*msg_buffer, *cmd, *node_id, *variable, *value, *new_name;
	int					rc;
	char				*saved_buffer;
	__uint32_t	len;

	KNOWN_UNUSED_PARAM(w_fd);

	if ((rc = read_message(r_fd, &msg_buffer, &len)) < 0)
		{
		printf("Slave unable to read msg from thread (%s)\n",strerror((-rc)));
		return;
		}
	saved_buffer = malloc(len);
	strcpy(saved_buffer, msg_buffer);

//	printf("Slave got msg (%s) from master\n",msg_buffer);
	if (!(cmd = strtok(msg_buffer, ":")))
		goto out;
	if (!(node_id = strtok(NULL, ":")))
		goto out;
	if (!strcasecmp(cmd, "NODE_NEW"))
		{
		if (!context_create(node_id))
			{
			printf("RSH Slave unable to create node context\n");
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok(NULL,":")))
			goto out;
		if (!(value = strtok(NULL,":")))
			goto out;
		if ((rc = context_variable_set(node_id, variable, value)) < 0)
			{
			printf("RSH Slave unable to set context variable %s (%s)\n",
						 variable,
						 strerror(-rc));
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char	*client_fd;
		int		fd;

		if (!(client_fd = strtok(NULL,":")))
			goto out;
		fd = atoi(client_fd);
		context_notify_disconnection_all(fd);
		goto out;
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		if (!(new_name = strtok(NULL,":")))
			goto out;
		if ((rc = context_rename(node_id, new_name)) < 0)
			{
			printf("RSH Slave unable to rename context %s (%s)\n",
						 node_id,
						 strerror(-rc));
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "NODE_DEL"))
		{
		if ((rc = context_destroy(node_id)) < 0)
			{
			printf("RSH Slave unable to destroy node context (%s)\n",strerror(-rc));
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "NODE_IPC"))
		{
		if ((rc = context_send_ipc(node_id, saved_buffer)) < 0)
			{
			printf("RSH Slave unable to send IPC (%s)\n",strerror(-rc));
			goto out;
			}
		}
	else if (!strcasecmp(cmd,"SLAVE_SHUTDOWN"))
		{
		context_destroy_all();
		exit(0);
		}
	else
		printf("RSH Slave received unknown msg from master (%s)\n",cmd);
out:
	free(msg_buffer);
	free(saved_buffer);
}

static void process_msg_from_thread(CONTEXT *ctx, int read_fd, int write_fd)
{
	char				*msg_buffer;
	int					rc;
	__uint32_t	len;

  KNOWN_UNUSED_PARAM(read_fd);
	if ((rc = read_message(ctx->slave_ipc_read, &msg_buffer, &len)) < 0)
		{
		printf("Slave unable to read msg from thread (%s)\n",strerror((-rc)));
		return;
		}
	// printf("Slave got msg (%s) from thread\n",msg_buffer);
	/* Forward message to master */
	send_message(write_fd, msg_buffer);
	free(msg_buffer);
}

static int read_message(int fd, char **buffer, __uint32_t *len)
{
	__uint32_t	net_len;
	int					rc;
	__uint32_t	brtr;
	char				*offset;

	brtr = sizeof(net_len);
	offset = (char *) &net_len;
	while(brtr)
		{
// printf("IPC length hunk read basep=%p, offset=%p, brtr=%d\n",&net_len,offset,brtr);
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
	*len = ntohl(net_len);

// printf("IPC length size =%d\n",*len);
	if (!(*buffer = malloc(*len)))
		{
		printf("Unable to allocate memory of size %d (0x%x)\n",*len,*len);
		return(-ENOMEM);
		}
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
// printf("IPC data hunk read basep=%p, offset=%p, brtr=%d\n",*buffer,offset,brtr);
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset +=rc;
		}
// printf("IPC read complete (msg %s)\n",*buffer);
	return(0);
}

static int send_message(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}

