
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zacs@valinux.com>.
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2001, San Mehat, Zac Sprackett, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <sbt2_module_harness.h>
#include <sbt2_slave_context.h>
#include <prototypes.h>
#include <sbt2_slave_context.h>
#include <pthread.h>
#include <tupelo.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_sbt2_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static int	sbt2_engage(CONTEXT *);
static void	sbt2_disengage(CONTEXT *);
static int  sbt2_reengage(CONTEXT *);
static int sbt2_status(CONTEXT *);
static void set_device_address(CONTEXT *, char *);

/* SBT2 forward declarations */
static u16 pppfcs16(register u16, register unsigned char *, register int);
#ifdef DUMP_DATA
static void dump_data(char *, unsigned char *, int);
#endif
static int	create_icmb(CONTEXT *,
									unsigned char	*,
							 		unsigned char,
									unsigned char,
									unsigned char,
									unsigned char,
									unsigned char,
									unsigned char	*,
									int);
static int sbt2_sequence_init(CONTEXT *);
static int	sbt2_power_off(CONTEXT *);
static int	sbt2_power_on(CONTEXT *);
static int	sbt2_hard_reset(CONTEXT *);
static int do_trans(CONTEXT *,
						unsigned char *,
						int,
						unsigned char	*,
						int	*,
						int);
static VAR_BIND *varbind_locate(unsigned char *, int);
static int unnormalize_data(unsigned char *, unsigned char *,int);
static int normalize_data(unsigned char *, unsigned char *);

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
	sbt2_disengage(ctx);
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			continue;
			}
		if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		pthread_testcancel();
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_sbt2_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;

		/* BEGIN San mod 07/12/00 */
		value = &saved_buffer[((variable - msg_buffer)+(strlen(variable)+1))];
		/* END San mod 07/12/00 */
		if (!strcasecmp(variable,"DEVICE_ADDRESS"))
			{
			set_device_address(ctx, value);
			sbt2_disengage(ctx);

			/* If we have a device, we should engage */
			if (!strcasecmp(value, "NONE"))
				sbt2_disengage(ctx);
			else
				{
				if ((rc = sbt2_engage(ctx)) < 0)
					{
					return;
					}
				}
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char								*fd;
		int									client_fd;

		/* Called when a client disconnects from nexxus */
		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
		goto out;
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	*new_name;
		char  *tmp;
	
		/* Called when a node is renamed */
		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;

		tmp = ctx->node_id;
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		free(tmp);
		goto out;
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Process an IPC message
 */
static void process_sbt2_ipc(CONTEXT			*ctx,
															__uint32_t	job_id,
															int					client_fd,
															char				*ipc_cmd,
															char				*ipc_args)
{
	char					resp_msg[255];
	char					*last = (char *) 0x00;
	int						rc;

	if (!strcasecmp(ipc_cmd, "CONFIGURATION"))
		{
		CONTEXT *ctx_scan;
		char *dev;

		if (!(dev = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (strcasecmp(dev,"NONE"))
			{
			/* Scan for device in use */
			pthread_mutex_lock(&context_mutex);
			ctx_scan = context_head;
			while(ctx_scan)
				{
				if (ctx_scan->device_address)
					if ((!strcmp(ctx_scan->device_address,dev)) &&
						(ctx_scan != ctx))
						break;
					ctx_scan = ctx_scan->next;
				}
			pthread_mutex_unlock(&context_mutex);
			if (ctx_scan)
				{
				sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR,%s",
					job_id,
					client_fd,
					strerror(EADDRINUSE));
				send_msg(ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			}

		/* See if the serial port has changed */
		if ((ctx->device_address ? strcmp(ctx->device_address, dev) : 1))
			{
			sprintf(resp_msg,
				"0:0:SET:%s:SBT2:DEVICE_ADDRESS:%s",
				ctx->node_id,
				dev);
			send_msg(ctx->thread_ipc_write,resp_msg);
			set_device_address(ctx,dev);

			/* Reengage on the new serial port */
			sbt2_reengage(ctx);
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	/* Anything below this requires a configured device */
	if (!ctx->device_address)
		{
		sprintf(resp_msg,
			"%d:%d:FOR:JOB_ERROR:%s",
		job_id, client_fd, strerror(ENOPKG));
		send_msg(ctx->thread_ipc_write,resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "REFRESH"))
		{
		sbt2_reengage(ctx);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "INQUIRY"))
		{
		sprintf(resp_msg,
			"%d:%d:FOR:INQUIRY:%s:%s",
			job_id,client_fd,ctx->node_id,ctx->device_address);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

	/* All of these require a connected device */
	if (ctx->con_sock == -1)
		{
		sprintf(resp_msg,
			"%d:%d:FOR:JOB_ERROR:%s",
		job_id, client_fd, strerror(ENOTCONN));
		send_msg(ctx->thread_ipc_write,resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "POWER_ON"))
		{
		if ((rc = sbt2_power_on(ctx))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(-rc));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "POWER_OFF"))
		{
		if ((rc = sbt2_power_off(ctx))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(-rc));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "HARD_RESET"))
		{
		if ((rc = sbt2_hard_reset(ctx))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(-rc));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "CHASSIS_STATUS"))
		{
		if ((rc = sbt2_status(ctx))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(-rc));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		switch((unsigned char)rc)
			{
			case 0x00:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:WORKING",
					job_id,client_fd);
				break;
			case 0x03:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:SUSPEND_TO_RAM",
					job_id,client_fd);
				break;
			case 0x04:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:SUSPEND_TO_DISK",
					job_id,client_fd);
				break;
			case 0x05:
			case 0x06:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:SOFT_OFF",
					job_id,client_fd);
				break;
			case 0x07:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:MECHANICAL_OFF",
					job_id,client_fd);
				break;
			case 0x08:
			case 0x09:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:SLEEPING",
					job_id,client_fd);
				break;
			case 0x20:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:LEGACY_ON",
					job_id,client_fd);
				break;
			case 0x21:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:LEGACY_OFF",
					job_id,client_fd);
				break;
			case 0x2a:
			default:
				sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:UNKNOWN",job_id,client_fd);
			}
		send_msg(ctx->thread_ipc_write, resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}
	
/*
 * Receive a message. This function is threadsafe
 */
static int recv_msg(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
	char				*offset;
  int         rc;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
	else if (!rc)
		return(-1);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int send_msg(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

	if (fd == -1)
		return(-EBADF);
  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}

static int sbt2_engage(CONTEXT *ctx)
{
	char *last = (char *) 0x00;
	char log_msg[255];
#ifndef __sgi
	struct serial_struct  ser;
#endif /* __sgi */

	if (ctx->con_sock != -1)
		return(-EADDRINUSE);

	if (!ctx->device_address)
		return(-EFAULT);

	/* Is this a serial port or a network address */
	if (ctx->device_address[0] == '/')
		{
		struct termios tty;
		if ((ctx->con_sock = open(ctx->device_address, O_RDWR))<0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s failed to engage on serial port (%m)",
				ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		tcgetattr(ctx->con_sock, &tty);
		cfmakeraw(&tty);
		cfsetospeed(&tty,B19200);
		cfsetispeed(&tty,B19200);
		tcsetattr(ctx->con_sock,TCSANOW,&tty);
#ifndef __sgi
		if (ioctl (ctx->con_sock, TIOCGSERIAL, &ser) == 0)
			{
			if (!(ser.flags & ASYNC_LOW_LATENCY))
				{
				ser.flags |= ASYNC_LOW_LATENCY;
				ioctl (ctx->con_sock, TIOCSSERIAL, &ser);
				}
			}
#endif  /*__sgi*/
		}
	else
		{
		struct sockaddr_in sock;
		int optval;
		char address[64];
		unsigned int port;
		char *bword;
		char tmp[64];

		strcpy(tmp, ctx->device_address);

		if (!(bword = strtok_r(tmp, "-", &last)))
			return(-EFAULT);
		strcpy(address,bword);

		if (!(bword = strtok_r(NULL, "-", &last)))
			return(-EFAULT);

		port = atoi(bword);

		if ((ctx->con_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		optval = 1;
		if (setsockopt(ctx->con_sock,SOL_SOCKET,SO_REUSEADDR,
			&optval,sizeof(optval))<0)
				{
				sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
				send_msg(ctx->thread_ipc_write, log_msg);
				return(-errno);
				}

		memset(&sock, 0, sizeof(sock));
		sock.sin_family = AF_INET;
		sock.sin_addr.s_addr = inet_addr(address);
		sock.sin_port = htons(port);

		if (connect(ctx->con_sock,(struct sockaddr *) &sock,sizeof(sock)) < 0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		}
	sprintf(log_msg,"0:0:LOG:Thread %s engage successful",ctx->node_id);
	send_msg(ctx->thread_ipc_write, log_msg);
	if (sbt2_sequence_init(ctx)<0)
		{
		sbt2_disengage(ctx);
		return(-1);
		}

	return(0);
}

static void sbt2_disengage(CONTEXT *ctx)
{
	if (ctx->con_sock != -1)
		{
		char log_msg[255];
		int  tmp_sock;

		sprintf(log_msg, "0:0:LOG:Thread %s disengaging from %s",
			ctx->node_id, ctx->device_address);
		send_msg(ctx->thread_ipc_write, log_msg);

		tmp_sock = ctx->con_sock;
		ctx->con_sock = -1;
		close(tmp_sock);
		ctx->next_ssn = 0x00000000;
		}
}

static int sbt2_reengage(CONTEXT *ctx)
{
	sbt2_disengage(ctx);
	return(sbt2_engage(ctx));
}

//
// hard_reset the unit. Returns 0 on success, -errno on failure
//
static int	sbt2_hard_reset(CONTEXT *ctx)
{
	int						resp_len, ipmb_len, rc;
	unsigned char	ipmb_req[255];
	unsigned char	ipmb_resp[32];
	char          log_msg[255];
	unsigned char	arg;

	memset(ipmb_req, 0, sizeof(ipmb_req));
  arg = 3;
  ipmb_len = create_icmb(ctx,
												ipmb_req,
                        CMD_CHASSIS_CONTROL,
                        CHASSIS_REQUEST,
                        0x00,
                        SBT2_SLAVE_ADDRESS,
                        SBT2_APPLICATION_SLAVE_ADDRESS,
                        &arg,
                        1);
	resp_len = sizeof(ipmb_resp);
	while (1)
		{
  	if ((rc = do_trans(ctx, ipmb_req, ipmb_len, ipmb_resp, &resp_len, 0))<0)
			{
			if (rc == -EAGAIN)
				continue;
			sprintf(log_msg, "0:0:LOG:do_trans() failed. (%s)",
				strerror(-rc));
			send_msg(ctx->thread_ipc_write, log_msg);
			return(rc);
			}
		break;
		}
	if (ipmb_resp[7] != 0x00)
		{
		sprintf(log_msg, "0:0:LOG:Power off failed. (cc=0x%.2x)",
			ipmb_resp[7]);
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-EIO);
		}
	return(0);
}

//
// Power off the unit. Returns 0 on success, -errno on failure
//
static int	sbt2_power_off(CONTEXT *ctx)
{
	int						resp_len, ipmb_len, rc;
	unsigned char	ipmb_req[255];
	unsigned char	ipmb_resp[32];
	unsigned char	arg;
	char          log_msg[255];

	memset(ipmb_req, 0, sizeof(ipmb_req));
  arg = 0;
  ipmb_len = create_icmb(ctx,
												ipmb_req,
                        CMD_CHASSIS_CONTROL,
                        CHASSIS_REQUEST,
                        0x00,
                        SBT2_SLAVE_ADDRESS,
                        SBT2_APPLICATION_SLAVE_ADDRESS,
                        &arg,
                        1);
	resp_len = sizeof(ipmb_resp);
	while (1)
		{
  	if ((rc = do_trans(ctx, ipmb_req, ipmb_len, ipmb_resp, &resp_len, 0))<0)
			{
			if (rc == -EAGAIN)
				continue;
			sprintf(log_msg, "0:0:LOG:do_trans() failed. (%s)",
				strerror(-rc));
			send_msg(ctx->thread_ipc_write, log_msg);
			return(rc);
			}
		break;
		}
	if (ipmb_resp[7] != 0x00)
		{
		sprintf(log_msg, "0:0:LOG:Power off failed. (cc=0x%.2x)",
			ipmb_resp[7]);
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-EIO);
		}
	return(0);
}

//
// Power on the unit. Returns 0 on success, -errno on failure
//
static int	sbt2_power_on(CONTEXT *ctx)
{
	int						resp_len, ipmb_len, rc;
	unsigned char	ipmb_req[255];
	unsigned char	ipmb_resp[32];
	char					log_msg[255];
	unsigned char	arg;

	memset(ipmb_req, 0, sizeof(ipmb_req));
  arg = 1;
  ipmb_len = create_icmb(ctx,
												ipmb_req,
                        CMD_CHASSIS_CONTROL,
                        CHASSIS_REQUEST,
                        0x00,
                        SBT2_SLAVE_ADDRESS,
                        SBT2_APPLICATION_SLAVE_ADDRESS,
                        &arg,
                        1);
	resp_len = sizeof(ipmb_resp);
  if ((rc = do_trans(ctx, ipmb_req, ipmb_len, ipmb_resp, &resp_len, 0))<0)
		{
		sprintf(log_msg, "0:0:LOG:do_trans() failed. (%s)",
			strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		return(rc);
		}

	if (ipmb_resp[7] != 0x00)
		{
		sprintf(log_msg, "0:0:LOG:Power off failed. (cc=0x%.2x)",
			ipmb_resp[7]);
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-EIO);
		}
	return(0);
}

static int sbt2_status(CONTEXT *ctx)
{
	int						resp_len, ipmb_len, rc;
	unsigned char	ipmb_req[255];
	unsigned char	ipmb_resp[32];
	char					log_msg[255];
	POWER_STATE_RESPONSE *power_resp;

	memset(ipmb_req,0,sizeof(ipmb_req));
	ipmb_len = create_icmb(ctx,
		ipmb_req,
		CMD_GET_CURRENT_POWER_STATE,
		APP_REQUEST,
		0x00,
		SBT2_SLAVE_ADDRESS,
		SBT2_APPLICATION_SLAVE_ADDRESS,
		NULL,
		0);
	resp_len = sizeof(ipmb_resp);
  if ((rc = do_trans(ctx, ipmb_req, ipmb_len, ipmb_resp, &resp_len, 0))<0)
		{
		sprintf(log_msg, "0:0:LOG:do_trans() failed. (%s)",
			strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		return(rc);
		}

	if (ipmb_resp[7] != 0x00)
		{
		sprintf(log_msg, "0:0:LOG:Chassis status failed. (cc=0x%.2x)",
			ipmb_resp[7]);
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-EIO);
		}

	power_resp = (POWER_STATE_RESPONSE *) &ipmb_resp[7];
	return((int)power_resp->sps_enumeration);
}


//
// Function executes a 'chassis status' command in order to figure out what
// to synchronize our sequence counter to. sets global variable 'next_ssn'
//
// Use this function FIRST to initialize the unit
// Returns 0 on success, -1 on error (no hardware detected, etc)
//

static int sbt2_sequence_init(CONTEXT *ctx)
{
	int						resp_len, ipmb_len, rc;
	unsigned char	ipmb_req[255];
	unsigned char	ipmb_resp[32];
	unsigned char	arg;
	char					log_msg[255];

	memset(ipmb_req, 0, sizeof(ipmb_req));
  arg = 1;
  ipmb_len = create_icmb(ctx,
												ipmb_req,
                        CMD_GET_CHASSIS_STATUS,
                        CHASSIS_REQUEST,
                        0x00,
                        SBT2_SLAVE_ADDRESS,
                        SBT2_APPLICATION_SLAVE_ADDRESS,
                        NULL,
                        0);
	resp_len = sizeof(ipmb_resp);
  if ((rc = do_trans(ctx, ipmb_req, ipmb_len, ipmb_resp, &resp_len, 1))<0)
		{
		sprintf(log_msg,"0:0:LOG:Thread %s failed to initialize (%s)",
			ctx->node_id,strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		return(rc);
		}
	return(0);
}

/*
 * Scan through a response packet and locate the variable bindings 
 * structure.. ugh.
 */
static VAR_BIND *varbind_locate(unsigned char *response_pkt, int len)
{
	int				i;
	VAR_BIND	*p;

	for (i=0; i < len; i++)
		{
		p = (VAR_BIND *) &response_pkt[i];
		if ((p->language == 0x19) &&
				(p->response_format == 0x77000000) &&
				(p->trap_src_type == 0x20) &&
				(p->event_src_type == 0x20))
			return(p);
		}
	return(NULL);
}

#ifdef DUMP_DATA
static void dump_data(char *name, unsigned char *data, int len)
{
	unsigned char	*p;
	int						i;

	printf("Data dump of [%s] (%d bytes)\n",name,len);

	for (p = data; ((p - data) < len);)
		{
		for (i=0; i < 16; i++, p++)
			{
			printf("%.2x ",*p);
			if ((p - data) == len)
				break;
			}
		printf("\n");
		}
}
#endif

//
// PPP frame checksum generation code
//
static u16 pppfcs16(fcs, cp, len)
    register u16 fcs; 
    register unsigned char *cp;
    register int len;
{ 
  while (len--)
    fcs = (fcs >>8) ^ fcstab[(fcs ^ *cp++) & 0xff];
  return (fcs);
}

//
// Create a remote control packet, send it, and get the response
// ctx      = CONTEXT
// ipmb_req = ipmb request to send
// ipmb_len = length of ipmb request
// syncflag = set if we're synchronizing (normally 0)
// ipmb_resp= ipmb response. User supplies buffer
// resp_len = max length of response buffer. Function modifies to be actual
//            response length on exit
// Successfull return is 0. anything else is errno
//
static int do_trans(CONTEXT *ctx,
						unsigned char *ipmb_req,
						int						ipmb_len,
						unsigned char	*ipmb_resp,
						int						*resp_len,
						int						syncflag)
{
	VAR_BIND				*variable_bind;
	TL							*tl;
	struct timeval	to;
	fd_set					read_fds;
	unsigned char		*response_buffer, *send_buffer, *ipmb_msg, *p, format_version;
	unsigned char		buff[4096], pre_amble[2], post_amble[2];
	int							pktlen, unnormal_len, rc, cnt = 0, varbind_offset;
	int							response_len;
	u16							fcs;
	struct iphdr		*ip;
	struct udphdr		*udp;
	struct bmchdr		*bmc;
	struct reqhdr		*req;
	unsigned int		*ssn;
	char						log_msg[255];

	send_buffer = malloc(4096);
	memset(send_buffer, 0x00, 4096);
	memset(&buff[0], 0, sizeof(buff));

#define PS 1 // packet skew

	buff[0]= 0x21;

	ip = (struct iphdr *) &buff[PS];
	udp = (struct udphdr *) &buff[PS+sizeof(struct iphdr)];
	bmc = (struct bmchdr *) &buff[PS+sizeof(struct iphdr)+
																					sizeof(struct udphdr)];
	req = (struct reqhdr *) &buff[PS+sizeof(struct iphdr)+
																					sizeof(struct udphdr)+
																					sizeof(struct bmchdr)];
	ipmb_msg = (unsigned char *) &buff[PS+sizeof(struct iphdr)+
																					sizeof(struct udphdr)+
																					sizeof(struct bmchdr)+
																					sizeof(struct reqhdr)];

	udp->dest = htons(623);
	bmc->field1 = 0x06;
	bmc->hdr_len = 0x20;
	bmc->seq = (htonl(ctx->next_ssn) & 0xff);
	bmc->res_constant = 0x77000000;
	bmc->field2 = 0x08;
	bmc->field_version = 0x01;
	bmc->sati = 0x01;
	bmc->ssn = ctx->next_ssn;

	memset(bmc->md5, 0x00, sizeof(bmc->md5));
	req->msg_type = 0x90;
	req->msg_length = ipmb_len;
	memcpy(ipmb_msg, ipmb_req, ipmb_len);
	pktlen = (ipmb_msg - &buff[0])+ ipmb_len ;

	fcs = pppfcs16(0xffff, &buff[0], pktlen);
	fcs ^=0xffff;
	buff[(ipmb_msg - &buff[0])+ipmb_len] = (fcs & 0x00ff);
	buff[(ipmb_msg - &buff[0])+ipmb_len+1] = ((fcs >>8) & 0x00ff);
	pktlen+=2;

	// now encode
	unnormal_len = unnormalize_data(&buff[0], send_buffer, pktlen);

#if DEBUG
	sprintf(log_msg, "0:0:LOG:[create] SSN = 0x%.8x (%d), bmc seq = %x (%d)\n",
									ntohl(bmc->ssn),
									ntohl(bmc->ssn),
									bmc->seq,
									bmc->seq); 
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	/* Now send it */
	pre_amble[0]= 0x1b;
	pre_amble[1] = 0x7e;
	post_amble[0] = 0x7e;
	post_amble[1] = 0x1b;

	write(ctx->con_sock, &pre_amble, 2);
	write(ctx->con_sock, send_buffer,unnormal_len);
	write(ctx->con_sock, &post_amble, 2);
	/* Done with buffer... free it */
	free(send_buffer);

	/* Now wait for response */
	while(1)
		{
		to.tv_sec = 1;
		to.tv_usec = 0;
		FD_ZERO(&read_fds);
		FD_SET(ctx->con_sock, &read_fds);

		if ((rc = select ((ctx->con_sock +1), &read_fds, NULL, NULL, &to))<0)
			{
			sprintf(log_msg,"0:0:LOG:Error in select. (%m)\n");
			send_msg(ctx->thread_ipc_write,log_msg);
			return(-errno);
			}
		if (!rc)
			break;
		else
			{
			if ((rc = read(ctx->con_sock, &buff[cnt], (sizeof(buff)-cnt)))<0)
				{
				sprintf(log_msg,"0:0:LOG:Error in read. (%m)\n");
				send_msg(ctx->thread_ipc_write,log_msg);
				return(-errno);
				}
			cnt +=rc;
			}
		}
	if (cnt == 0)
		{
		return (-EAGAIN);
		}
	response_buffer = malloc(cnt);
	response_len = normalize_data(&buff[0], response_buffer);

	if (!(variable_bind = varbind_locate(&response_buffer[0], response_len)))
		{
		free(response_buffer);
		sprintf(log_msg,"0:0:LOG:Unable to find varbind.\n");
		send_msg(ctx->thread_ipc_write,log_msg);
		return(-EIO);
		}
	varbind_offset = ((void *) variable_bind - (void *) &response_buffer[0]);

	ssn = (unsigned int *) &variable_bind->event_data2;
	ctx->next_ssn = *ssn;

	if (variable_bind->event_data1 == 0x02)
		{
		free(response_buffer);
		if (syncflag)
			return(0);
		return(-EAGAIN);
		}
	else if (variable_bind->event_data1 != 0x00)
		{
		free(response_buffer);
		sprintf(log_msg,"0:0:LOG:Event data 1 0x%x\n",variable_bind->event_data1);
		send_msg(ctx->thread_ipc_write,log_msg);
		return(-EIO);
		}
	if (syncflag)
		{
		free(response_buffer);
		*resp_len = 0;
		return(0);
		}
	// Copy the ICMB response for the user
	p = (unsigned char *) &response_buffer[varbind_offset+(sizeof(VAR_BIND))];
	tl = (TL *) p++;
	format_version = *p;
	if (format_version != 0x01)
		{
		sprintf(log_msg,"0:0:LOG:BAD ICMB response data (fmt version = 0x%x)\n",
			format_version);
		send_msg(ctx->thread_ipc_write,log_msg);
		free(response_buffer);
		return (-EIO);
		}
	if (tl->num_bytes > *resp_len)
		{
		free(response_buffer);
		return(-EOVERFLOW);
		}
	*resp_len = tl->num_bytes;
	memcpy(ipmb_resp, p, tl->num_bytes);
	free(response_buffer);
	return(0);
}

static int unnormalize_data(unsigned char *src, unsigned char *dst,int I_srclen)
{
  int I_srcpos,I_dstpos;
  
  for (I_srcpos=0,I_dstpos=0;I_srcpos<I_srclen;I_srcpos++)
    {
    if (src[I_srcpos]== (unsigned char) 0x7e)
      {
      dst[I_dstpos++]= (unsigned char) 0x7d;
      dst[I_dstpos++]= (unsigned char) 0x5e;
      }
    else if (src[I_srcpos]== (unsigned char) 0x7d)
      {
      dst[I_dstpos++]= (unsigned char) 0x7d;
      dst[I_dstpos++]= (unsigned char) 0x5d;
      }
    else if (src[I_srcpos]== (unsigned char) 0x11)
      {
      dst[I_dstpos++]= (unsigned char) 0x7d;
      dst[I_dstpos++]= (unsigned char) 0x31;
      }
    else if (src[I_srcpos]== (unsigned char) 0x13)
      {
      dst[I_dstpos++]= (unsigned char) 0x7d;
      dst[I_dstpos++]= (unsigned char) 0x33;
      }
    else
      dst[I_dstpos++]=src[I_srcpos];
    } 
  return(I_dstpos);
}

static int normalize_data(unsigned char *src, unsigned char *dst)
{
  int I_c1,I_c2,I_pktlen;
  
  I_pktlen = 0;
  for (I_c1=1,I_c2=0;src[I_c1]!=(unsigned char) 0x7e;I_c1++)
    {
    if (I_c1>255)
      return(-EFAULT);
    if (src[I_c1]== (unsigned char) 0x7d)
      {
			if (src[I_c1+1] == 0x5e)
				dst[I_c2] = 0x7e;
			if (src[I_c1+1] == 0x5d)
				dst[I_c2] = 0x7d;
			if (src[I_c1+1] == 0x31)
				dst[I_c2] = 0x11;
			if (src[I_c1+1] == 0x33)
				dst[I_c2] = 0x13;
      I_c2++;
      I_c1++;
      }
    else
      {
      dst[I_c2]=src[I_c1];
      I_c2++;
      }
    } 
  dst[I_c2]=src[I_c1];
  I_pktlen = I_c2;
  return(I_pktlen);
}

//
// Code to generate an icmb packet
//
static int		create_icmb(CONTEXT *ctx,
									unsigned char	*buffer,
							 		unsigned char	cmd,
									unsigned char	netfn,
									unsigned char	lun,
									unsigned char	rs_slave_addr,
									unsigned char	rq_slave_addr,
									unsigned char	*data,
									int						dlen)
{
	IPMB_REQHDR		*hdr = (IPMB_REQHDR *) buffer;
	unsigned char	cksum;
	int						i;
	int						icmb_size;

	hdr->rs_slave_addr	= rs_slave_addr;
	hdr->net_fn					= netfn;
	hdr->lun						= lun;
	hdr->cksum1					= ((0xff - hdr->rs_slave_addr -
												(*(unsigned char *) &buffer[1]))+1);
	hdr->rq_slave_addr 	= rq_slave_addr;
	hdr->rqseq					= ctx->icmb_seq+=4;
	hdr->cmd						= cmd;

	if (dlen)
		memcpy(&buffer[sizeof(IPMB_REQHDR)], data, dlen);
	cksum = 0xff - hdr->rq_slave_addr - hdr->rqseq - hdr->cmd;
	for (i = sizeof(IPMB_REQHDR); i < (int)(sizeof(IPMB_REQHDR)+dlen); i++)
		cksum = cksum - buffer[i];
	buffer[(sizeof(IPMB_REQHDR)+dlen)] = cksum+1;
	icmb_size = sizeof(IPMB_REQHDR) + dlen + 1;
	return(icmb_size);
}

static void set_device_address(CONTEXT *ctx, char *dev)
{
	if (ctx->device_address)
		free(ctx->device_address);
	ctx->device_address = malloc((strlen(dev)+1));
	strcpy(ctx->device_address,dev);
}
