/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * VACM Remote System Status Module
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <libloose.h>
#include <sys_stat.h>

#include <config.h>

#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

static void			config_cb(__uint32_t job_id, int client_fd, char *msg);
static void			deinit_cb(__uint32_t job_id, int client_fd, char *msg);
static void			discon_cb(__uint32_t job_id, int client_fd, char *msg);
static void			ipc_cb(char				*node_id,
											 __uint32_t	job_id,
											 int				client_fd,
											 char				*msg);
static int			do_request(int				request,
													 NODELIST		*node,
													 __uint32_t	job_id,
													 int				client_fd);
static int			do_disconnect(NODELIST *node);
static int			do_connect(NODELIST	*node, __uint32_t job_id, int client_fd);
static char			*return_req_str(int	request);
static NODELIST	*node_create(char *node_id);
static NODELIST	*lookup_node(char *node_id);
static void			node_destroy(NODELIST	*node);
static int			do_auth(NODELIST *node);
static int			do_proxy_auth(NODELIST *node);
static int			do_proxy_connect(NODELIST *node);

extern int errno;

static NODELIST	*nodelist_head = NULL;
static NODELIST	*nodelist_tail = NULL;

#ifdef OPENSSL
static	SSL_CTX			*ctx = NULL;
static	SSL_METHOD	*meth = NULL;
#endif

int main(argc, argv)
int		argc;
char	**argv;
{
	KNOWN_UNUSED_PARAM(argc);
	KNOWN_UNUSED_PARAM(argv);

	lm_init();
	lm_nexxus_connect();
	if (lm_register("Sysstat",
									"System Status Daemon Module",
									"Module to interface with a remote VACM System Status Daemon",
									"San Mehat (nettwerk@valinux.com)",
									"SYSSTAT",
									0,
									1) < 0)
		{
		printf("Unable to register module with Nexxus\n");
		exit(-1);
		}

#ifdef OPENSSL
  SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
  meth = SSLv2_client_method();
  ctx = SSL_CTX_new(meth);
	if (!ctx)
		{
		ERR_print_errors_fp(stderr);
		exit(0);
		}
#endif

	lm_main_loop(config_cb, deinit_cb, discon_cb, ipc_cb);
	exit(0);
}

static void config_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;
	char			*cmd_module;
	char			*node_id;
	char			logmsg[255];

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	cmd_module = strtok(msg, ":");
	node_id = strtok(NULL, ":");

	node = lookup_node(node_id);

	if (!strcasecmp(cmd_module, "NODE"))
		node = node_create(node_id);
	else if (!strcasecmp(cmd_module, "END_NODE"))
		{
		}
	else if (!strcasecmp(cmd_module, "ADDITION"))
		{
		node = node_create(node_id);
		return;
		}
	else if (!strcasecmp(cmd_module, "DELETION"))
		{
		node_destroy(node);
		return;
		}
	else if (!strcasecmp(cmd_module, "RENAME"))
		{
		char	*new_name;
		
		if (!(new_name = strtok(NULL,":")))
			{
			printf("[SYSSTAT] Malformatted cfg msg\n");
			return;
			}
		strcpy(node->node_id, new_name);
		return;
		}
	else if (!strcasecmp(cmd_module, "GLOBAL"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			printf("[SYSSTAT] Malformatted cfg msg\n");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			printf("[SYSSTAT] Malformatted cfg msg\n");
			return;
			}
		/* Any global variabls we care about here... ie: */
		if (!strcasecmp(variable, "IP_ADDRESS"))
			{
			node->ip_address = inet_addr(value);
			}
		}
	else if (!strcasecmp(cmd_module, "SYSSTAT"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			printf("[SYSSTAT] Malformatted cfg msg\n");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			printf("[SYSSTAT] Malformatted cfg msg\n");
			return;
			}
		/* Process any local variables here */
		if (!strcasecmp(variable,"AUTH_PASSWORD"))
			{
			strncpy(node->password, value, (sizeof(node->password)-1));
			return;
			}
		else if (!strcasecmp(variable,"PROXY_ADDR"))
			{
			strncpy(node->proxy_addr, value, (sizeof(node->proxy_addr)-1));
			return;
			}
		else if (!strcasecmp(variable,"PROXY_PASSWORD"))
			{
			strncpy(node->proxy_passwd, value, (sizeof(node->proxy_passwd)-1));
			return;
			}
		sprintf(logmsg,"Received unknown local module variable '%s'\n",
					variable);
		lm_log(logmsg);
		}
	else
		{
		sprintf(logmsg,"Received unknown config event <%s>",cmd_module);
		lm_log(logmsg);
		}
	if ((node->ip_address) && (node->password[0]))
		{
		struct in_addr	addr;
		addr.s_addr = node->ip_address;
#ifdef DEBUG
		sprintf(logmsg,"Node '%s' (ip %s) ready",
					 node_id,inet_ntoa(addr));
		lm_log(logmsg);
#endif
		}
}

static void deinit_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);

	lm_nexxus_disconnect();
	exit(0);
}

static void discon_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
  KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);
}

static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg)
{
	char			*cmd;
	char			response_msg[255];
	NODELIST	*node;
	char			*toss;
	int				rc = 0;


	if (!(cmd = strtok(msg, ":")))
		goto out_malformatted;

	if (!(toss = strtok(NULL, ":")))
		goto out_malformatted;

	if (!(node = lookup_node(node_id)))
		{
		sprintf(response_msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_FOUND",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		return;
		}
	/* Process IPC here */
	if (!strcasecmp(cmd, "CONFIGURATION"))
		{
		char	*daemon_password;
		int		i;
	
		if (!(daemon_password = strtok(NULL,":")))
			{
      sprintf(response_msg,
              "%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
              job_id,
              client_fd);
      lm_send_to_nexxus(response_msg);
      return;
			}
		if ((int)strlen(daemon_password) > 40)
			{
      sprintf(response_msg,
              "%d:%d:FOR:JOB_ERROR:PASSWORD_TOO_LONG",
              job_id,
              client_fd);
      lm_send_to_nexxus(response_msg);
      return;
			}
    for (i=0; (unsigned int)i < strlen(daemon_password); i++)
      if (!isprint((int)daemon_password[i]))
        {
        sprintf(response_msg,
                "%d:%d:FOR:JOB_ERROR:BAD_PASSWORD",
                job_id,
                client_fd);
        lm_send_to_nexxus(response_msg);
        return;
        }
    sprintf(response_msg,
            "0:0:SET:%s:SYSSTAT:AUTH_PASSWORD:%s",
            node_id,
            daemon_password);
    lm_send_to_nexxus(response_msg);
    sprintf(response_msg,
            "%d:%d:FOR:JOB_COMPLETED",
            job_id,
            client_fd);
    lm_send_to_nexxus(response_msg);

		strcpy(node->password,daemon_password);
		if ((node->ip_address) && (node->password[0]))
			{
			struct in_addr	addr;
			addr.s_addr = node->ip_address;

			sprintf(response_msg,"Node '%s' (ip %s) ready",
					 		node_id,inet_ntoa(addr));
			lm_log(response_msg);
			}
    return;
		}
	if ((!node->ip_address) || (!node->password[0]))
		{
		sprintf(response_msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_CONFIGURED",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		return;
		}
	if (!strcasecmp(cmd, "MEM"))
		rc = do_request(REQUEST_MEM, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "SET_PROXY"))
		{
		char						*proxy_addr;
		char						*proxy_passwd;
		struct in_addr	proxy_a;
		int		i;

		if (!(proxy_addr= strtok(NULL,":")))
			{
      sprintf(response_msg,
              "%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
              job_id,
              client_fd);
      lm_send_to_nexxus(response_msg);
      return;
			}
		if (!(proxy_passwd= strtok(NULL,":")))
			{
      sprintf(response_msg,
              "%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
              job_id,
              client_fd);
      lm_send_to_nexxus(response_msg);
      return;
			}
		if ((int)strlen(proxy_passwd) > 40)
			{
      sprintf(response_msg,
              "%d:%d:FOR:JOB_ERROR:PASSWORD_TOO_LONG",
              job_id,
              client_fd);
      lm_send_to_nexxus(response_msg);
      return;
			}
    for (i=0; (unsigned int)i < strlen(proxy_passwd); i++)
      if (!isprint((int)proxy_passwd[i]))
        {
        sprintf(response_msg,
                "%d:%d:FOR:JOB_ERROR:BAD_PASSWORD",
                job_id,
                client_fd);
        lm_send_to_nexxus(response_msg);
        return;
        }
		if (!strcasecmp(proxy_addr, "NONE"))
			{
			strcpy(node->proxy_addr,"NONE");
			}
		else
			{
			if (!inet_aton(proxy_addr, &proxy_a))
				{
       	sprintf(response_msg,
               "%d:%d:FOR:JOB_ERROR:BAD_ADDRESS",
               job_id,
               client_fd);
       	lm_send_to_nexxus(response_msg);
      	 return;
				}
			strcpy(node->proxy_addr,proxy_addr);
			}
    sprintf(response_msg,
            "0:0:SET:%s:SYSSTAT:PROXY_ADDR:%s",
            node_id,
            proxy_addr);
    lm_send_to_nexxus(response_msg);
    sprintf(response_msg,
            "0:0:SET:%s:SYSSTAT:PROXY_PASSWORD:%s",
            node_id,
            proxy_passwd);
    lm_send_to_nexxus(response_msg);
		strcpy(node->proxy_passwd,proxy_passwd);
    sprintf(response_msg,
            "%d:%d:FOR:JOB_COMPLETED",
            job_id,
            client_fd);
    lm_send_to_nexxus(response_msg);
		do_disconnect(node);
    return;
		}
	else if (!strcasecmp(cmd, "CPU_LOAD"))
		rc = do_request(REQUEST_CPU, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "UPTIME"))
		rc = do_request(REQUEST_UPTIME, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "FS"))
		rc = do_request(REQUEST_FS, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "WHO"))
		rc = do_request(REQUEST_WHO, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "PS"))
		rc = do_request(REQUEST_PS, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "VERSION"))
		rc = do_request(REQUEST_VERSION, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "APM"))
		rc = do_request(REQUEST_APM, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "INTERRUPTS"))
		rc = do_request(REQUEST_INT, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "IOPORTS"))
		rc = do_request(REQUEST_IO, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "DMA"))
		rc = do_request(REQUEST_DMA, node, job_id, client_fd);
	else if (!strcasecmp(cmd, "SWAP"))
		rc = do_request(REQUEST_SWAP, node, job_id, client_fd);
	else
		{
		sprintf(response_msg,
						"%d:%d:FOR:JOB_ERROR:UNSUPPORTED_MESSAGE",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		return;
		}
	if (!rc)
		{
		/* Send job completed message */
		sprintf(response_msg,
						"%d:%d:FOR:JOB_COMPLETED",
						job_id,
						client_fd);
		lm_send_to_nexxus(response_msg);
		}
	return;
out_malformatted:
	sprintf(response_msg,
					"%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
					job_id,
					client_fd);
	lm_send_to_nexxus(response_msg);
	return;
}

static int do_proxy_connect(NODELIST *node)
{
  __uint32_t      netlen;
  __uint32_t      reqlen;
  char            msg[255];
  int             rc = 0;
  fd_set          read_fds;
  struct timeval  to;
  char            rspstr[512];
	struct in_addr	addr;

	memset(&addr, 0, sizeof(addr));
	addr.s_addr = node->ip_address;

  snprintf(msg, sizeof(msg), "PROXY_CONNECT:%s",inet_ntoa(addr));
  reqlen = strlen(msg)+1;
  netlen = htonl(reqlen);
#ifdef OPENSSL
  rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
  rc = SSL_write(node->ssl, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#else
	rc = write(node->comm_sock, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
	rc = write(node->comm_sock, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#endif	/*OPENSSL*/
  while(1)
    {
    FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
    to.tv_sec = 5;
    to.tv_usec= 0;
    if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
      return(-1);
    else if (!rc)
			{
			lm_log("Timed out waiting for PROXY_CONNECT reply from proxy");
      return(-1);
			}
    if (FD_ISSET(node->comm_sock, &read_fds))
      {
#ifdef	OPENSSL
      rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
      rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#else
			rc = read(node->comm_sock, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
			rc = read(node->comm_sock, (char *) rspstr,reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#endif	/*OPENSSL*/
		
      if (!strncmp(rspstr,"PROXY_CONNECT_FAIL",18))
				{
				char	nexxus_msg[255];
			
				sprintf(nexxus_msg,
								"Proxy for node %s failed to connect to daemon",
								node->node_id);
				lm_log(nexxus_msg);
        return(-2);
				}
      else if (!strncmp(rspstr,"PROXY_CONNECT_OK",16))
        return(0);
      return(-1);
      }
    else
      return(-1);
    } /* end while */
}
static int do_proxy_auth(NODELIST *node)
{
  __uint32_t      netlen;
  __uint32_t      reqlen;
  char            msg[255];
  int             rc = 0;
  fd_set          read_fds;
  struct timeval  to;
  char            rspstr[512];

  snprintf(msg, sizeof(msg), "PROXY_AUTH:%s",node->proxy_passwd);
  reqlen = strlen(msg)+1;
  netlen = htonl(reqlen);
#ifdef OPENSSL
  rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
  rc = SSL_write(node->ssl, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#else
	rc = write(node->comm_sock, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
	rc = write(node->comm_sock, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#endif	/*OPENSSL*/
  while(1)
    {
    FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
    to.tv_sec = 5;
    to.tv_usec= 0;
    if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
      return(-1);
    else if (!rc)
			{
			lm_log("Timed out waiting for PROXY_AUTH reply from proxy");
      return(-1);
			}
    if (FD_ISSET(node->comm_sock, &read_fds))
      {
#ifdef	OPENSSL
      rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
      rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#else
			rc = read(node->comm_sock, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
			rc = read(node->comm_sock, (char *) rspstr,reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#endif	/*OPENSSL*/
		
      if (!strncmp(rspstr,"AUTH_ERROR",10))
				{
				char	nexxus_msg[255];
			
				sprintf(nexxus_msg,
								"Authentication to node '%s' proxy failed",
								node->node_id);
				lm_log(nexxus_msg);
        return(-2);
				}
      else if (!strncmp(rspstr,"AUTH_OK",7))
        return(0);
      return(-1);
      }
    else
      return(-1);
    } /* end while */
}

static int do_auth(NODELIST *node)
{
  __uint32_t      netlen;
  __uint32_t      reqlen;
  char            msg[255];
  int             rc = 0;
  fd_set          read_fds;
  struct timeval  to;
  char            rspstr[512];

  snprintf(msg, sizeof(msg), "AUTH:%s",node->password);
  reqlen = strlen(msg)+1;
  netlen = htonl(reqlen);
#ifdef OPENSSL
  rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
  rc = SSL_write(node->ssl, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#else
	rc = write(node->comm_sock, (char *) &netlen, sizeof(netlen));
  if ((!rc) || (rc <0))
    return(-1);
	rc = write(node->comm_sock, msg, reqlen);
  if ((!rc) || (rc <0))
    return(-1);
#endif	/*OPENSSL*/
  while(1)
    {
    FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
    to.tv_sec = 5;
    to.tv_usec= 0;
    if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
      return(-1);
    else if (!rc)
			{
			lm_log("Timed out waiting for AUTH reply from daemon");
      return(-1);
			}
    if (FD_ISSET(node->comm_sock, &read_fds))
      {
#ifdef	OPENSSL
      rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
      rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#else
			rc = read(node->comm_sock, (char *) &netlen, sizeof(netlen));
      if ((!rc) || (rc <0))
        return(-1);
      reqlen = ntohl(netlen);
			rc = read(node->comm_sock, (char *) rspstr,reqlen);
      if ((!rc) || (rc <0))
        return(-1);
#endif	/*OPENSSL*/
      if (!strncmp(rspstr,"AUTH_ERROR",10))
				{
				char	nexxus_msg[255];
			
				sprintf(nexxus_msg,
								"Authentication to node '%s' daemon failed",
								node->node_id);
				lm_log(nexxus_msg);
        return(-2);
				}
      else if (!strncmp(rspstr,"AUTH_OK",7))
        return(0);
      return(-1);
      }
    else
      return(-1);
    } /* end while */
}

static int do_request(int					request,
											NODELIST		*node,
											__uint32_t	job_id,
											int					client_fd)
{
	__uint32_t			netlen;
	__uint32_t			reqlen;
	char						msg[255];
	int							rc;
	char						reqstr[512];
	char						rspstr[512];
	fd_set					read_fds;
	struct timeval	to;

retry:
	if (node->comm_sock == -1)
		{
		if ((rc = do_connect(node, job_id, client_fd)) != 0)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:CONNECTION_FAILURE:%s",
							job_id,
							client_fd,
							strerror(rc));
			lm_send_to_nexxus(msg);
			return(-1);
			}
		}
	snprintf(reqstr,
					 sizeof(reqstr),
					 "REQUEST:%s:%d",
					 return_req_str(request),job_id);
	reqlen = strlen(reqstr)+1;
	netlen = htonl(reqlen);
#ifdef OPENSSL
	rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
	if ((!rc) || (rc <0))
		{
		lm_log("Reconnecting to status daemon after disconnection");
		do_disconnect(node);
		goto retry;
		}
	rc = SSL_write(node->ssl, reqstr, reqlen);
	if ((!rc) || (rc <0))
		{
		lm_log("Reconnecting to status daemon after disconnection");
		do_disconnect(node);
		goto retry;
		}
#else
	rc = write(node->comm_sock, (char *) &netlen, sizeof(netlen));
	if (rc < 0)
		{
		lm_log("Reconnecting to status daemon after disconnection");
		do_disconnect(node);
		goto retry;
		}
	rc = write(node->comm_sock, reqstr, reqlen);
	if (rc < 0)
		{
		lm_log("Reconnecting to status daemon after disconnection");
		do_disconnect(node);
		goto retry;
		}
#endif
	while(1)
		{
		FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
		to.tv_sec = 5;
		to.tv_usec= 0;
		if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:%m",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return(-1);
			}
		else if (!rc)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:TIMED OUT",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return(-1);
			}
		if (FD_ISSET(node->comm_sock, &read_fds))
			{
      char *bword;
      __uint32_t remote_id;

#ifdef OPENSSL
			rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
			if ((!rc) || (rc <0))
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
#else
			if ((rc = read(node->comm_sock, (char *) &netlen, sizeof(netlen)))<=0)
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
#endif
			reqlen = ntohl(netlen);
#ifdef OPENSSL
			rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
			if ((!rc) || (rc <0))
				{
				printf("rc = %d\n",rc);
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
#else
			if ((rc = read(node->comm_sock, (char *) rspstr, reqlen))<=0)
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
#endif
      if (!(bword = strtok(rspstr,":")))
        return(-1);

      remote_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);      

      bword = bword + strlen(bword) + 1;

      if (remote_id == job_id)
        {
  			if (!strncmp(bword,"ERROR",5))
	  			{
	  			lm_log("Daemon reported error");
	  			sprintf(msg,
	  							"%d:%d:FOR:JOB_ERROR:INTERNAL_ERROR",
	  							job_id,
	  							client_fd);
	  			lm_send_to_nexxus(msg);
	  			return(-1);
	  			}
	  		else if (!strncmp(bword,"DONE",4))
	  			return(0);
  
  			/* Forward the response off */
   	   sprintf(msg,
  			 				"%d:%d:FOR:%s",
			  				job_id,
			  				client_fd,
			  				bword);
		   lm_send_to_nexxus(msg);
       }
		 }
		else
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:CONNECTION_FAILURE:MISC ERROR",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return(-1);
			}
		} /* end while */
}

/*
 * disconnect from the daemon 
 */
static int do_disconnect(NODELIST *node)
{
	if (node->comm_sock == -1)
		return(-1);
	close(node->comm_sock);
#ifdef OPENSSL
	SSL_shutdown(node->ssl);
#endif
	node->comm_sock = -1;
	return(0);
}

/*
 * connect() to the daemon on the socket
 */
static int do_connect(NODELIST	*node, __uint32_t job_id, int client_fd)
{
	struct in_addr			inaddr;
	struct sockaddr_in	sock;
	int									optval = 1;
	int									optlen;
	fd_set							write_fds;
	struct timeval			to;
	int									rc;
	char								auth_errmsg[255];
	int									proxy_flag = 0;

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	if ((!strcasecmp(node->proxy_addr, "NONE")) || (node->proxy_addr[0] == '\0'))
		{
#ifdef DEBUG
		char logmsg[255];
#endif

		memset(&inaddr, 0, sizeof(inaddr));
		inaddr.s_addr = node->ip_address;
#ifdef DEBUG
		sprintf(logmsg,"Connecting to daemon %s",inet_ntoa(inaddr));
		lm_log(logmsg);
#endif
		}
	else
		{
#ifdef DEBUG
		char	logmsg[255];
#endif

		proxy_flag = 1;
		inet_aton(node->proxy_addr, &inaddr);
#ifdef DEBUG
		sprintf(logmsg,"Connecting to proxy %s",node->proxy_addr);
		lm_log(logmsg);
#endif
		}
	if ((node->comm_sock = socket(AF_INET, SOCK_STREAM, 0)) <0)
		{
		sprintf(auth_errmsg,"socket() failed (%m)\n");
		node->comm_sock = -1;
		lm_log(auth_errmsg);
		return(errno);
		}
  if (setsockopt(node->comm_sock,
                 SOL_SOCKET,
                 SO_REUSEADDR,
                 &optval,
                 sizeof(optval)) <0)
		{
		sprintf(auth_errmsg,"setsockopt() failed (%m)\n");
		lm_log(auth_errmsg);
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
	if ((int)(node->sock_opts =fcntl(node->comm_sock, F_GETFL))<0)
		{
		close(node->comm_sock);
		node->comm_sock = -1;
		sprintf(auth_errmsg,"fcntl() failed (%m)\n");
		lm_log(auth_errmsg);
		return(errno);
		}
	if (fcntl(node->comm_sock, F_SETFL, O_NONBLOCK)<0)
		{
		sprintf(auth_errmsg,"fcntl() failed (%m)\n");
		lm_log(auth_errmsg);
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	memcpy(&sock.sin_addr, &inaddr, sizeof(inaddr));
	sock.sin_port = htons(NETWORK_PORT);

	if (connect(node->comm_sock,
							(struct sockaddr *) &sock,
							sizeof(struct sockaddr)) <0)
		{
		if (errno != EINPROGRESS)
			{
			sprintf(auth_errmsg,"connect() failed (%m)\n");
			lm_log(auth_errmsg);
			close(node->comm_sock);
			node->comm_sock = -1;
			return(errno);
			}
		}

	FD_ZERO(&write_fds);
	FD_SET(node->comm_sock, &write_fds);
	to.tv_usec=0;
	to.tv_sec=5;
	if ((rc = select((node->comm_sock+1), NULL, &write_fds, NULL, &to)) < 0)
		{
		sprintf(auth_errmsg,"select() failed (%m)\n");
		lm_log(auth_errmsg);
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
	if (rc == 0)
		{
		if (proxy_flag)
			lm_log("Connection to proxy timed out");
		else
			lm_log("Connection to daemon timed out");
		close(node->comm_sock);
		node->comm_sock = -1;
		return(ETIMEDOUT);
		}
	if (!FD_ISSET(node->comm_sock, &write_fds))
		{
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
	optlen = sizeof(optval);
	if (getsockopt(node->comm_sock,
								 SOL_SOCKET,
								 SO_ERROR,
								 &optval,
								 &optlen) <0)
		{
		sprintf(auth_errmsg,"getsockopt() failed (%m)\n");
		lm_log(auth_errmsg);
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
	if (optval != 0)
		{
		close(node->comm_sock);
		node->comm_sock = -1;
		return(optval);
		}

	/* Set the socket to be blocking again */
	if (fcntl(node->comm_sock, F_SETFL, node->sock_opts)<0)
		{
		sprintf(auth_errmsg,"fcntl() failed (%m)\n");
		lm_log(auth_errmsg);
		close(node->comm_sock);
		node->comm_sock = -1;
		return(errno);
		}
#ifdef OPENSSL
	if (!(node->ssl = SSL_new(ctx)))
		{
		close(node->comm_sock);
		node->comm_sock = -1;
		return(EFAULT);
		}
	SSL_set_fd(node->ssl, node->comm_sock);
	if ((rc = SSL_connect(node->ssl))<0)
		{
		if (proxy_flag)
			lm_log("Unable to establish encrypted connection to proxy");
		else
			lm_log("Unable to establish encrypted connection to daemon");
		ERR_print_errors_fp(stderr);
		do_disconnect(node);
		return(ENETRESET);
		}
	
  if (!(node->cert=SSL_get_peer_certificate(node->ssl)))
		{
		do_disconnect(node);
		return(ENETRESET);
    }
  X509_free(node->cert);
#endif
	if (proxy_flag)
		{
		/* Authenticate to proxy */
		if ((rc = do_proxy_auth(node))<0)
			{
			do_disconnect(node);
    	return(EACCES);
			}
		/* Tell proxy to connect to daemon */
		if ((rc = do_proxy_connect(node))<0)
			{
			do_disconnect(node);
    	return(EACCES);
			}
		}
	/* Authenticate to daemon */
  if ((rc = do_auth(node))<0)
    {
		do_disconnect(node);
    return(EACCES);
    }
  else if (rc == -2)
    {
		sprintf(auth_errmsg,"Unable to authenticate to daemon. Bad password");
		lm_log(auth_errmsg);
		do_disconnect(node);
    return(EACCES);
		}
	return(0);
}

static NODELIST	*node_create(char *node_id)
{
	NODELIST	*new;

	if (!(new = (NODELIST	*) malloc(sizeof(NODELIST))))
		return(0);
	memset(new, 0, sizeof(NODELIST));
	if (!nodelist_head)
		nodelist_head = nodelist_tail = new;
	else
		{
		nodelist_tail->next = new;
		new->prev = nodelist_tail;
		nodelist_tail = new;
		}
	new->comm_sock = -1;
	strcpy(new->node_id, node_id);
	return(new);
}

static NODELIST	*lookup_node(char *node_id)
{
	NODELIST	*scan;

	if (!node_id)
		return(NULL);
	scan = nodelist_head;
	while(scan)
		{
		if (!strcmp(node_id, scan->node_id))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

static void node_destroy(NODELIST	*node)
{
  if (node == nodelist_head)
    {
    nodelist_head = nodelist_head->next;
    if (nodelist_head)
      nodelist_head->prev = NULL;
    else
      nodelist_tail = NULL;
    } 
  else if (node == nodelist_tail)
    {
    nodelist_tail = nodelist_tail->prev;
    if (nodelist_tail)
      nodelist_tail->next = NULL;
    else
      nodelist_head = NULL;
    } 
  else
    {
    node->next->prev = node->prev;
    node->prev->next = node->next;
    }
  free(node);
}

static char *return_req_str(int	request)
{
	switch(request)
		{
		case REQUEST_MEM:
			return strdup("MEM");
		case REQUEST_CPU:
			return strdup("CPU");
		case REQUEST_UPTIME:
			return strdup("UPTIME");
		case REQUEST_FS:
			return strdup("FS");
		case REQUEST_WHO:
			return strdup("WHO");
		case REQUEST_PS:
			return strdup("PS");
		case REQUEST_VERSION:
			return strdup("VERSION");
		case REQUEST_APM:
			return strdup("APM");
		case REQUEST_INT:
			return strdup("INT");
		case REQUEST_IO:
			return strdup("IO");
		case REQUEST_DMA:
			return strdup("DMA");
		case REQUEST_SWAP:
			return strdup("SWAP");
		default:
			return strdup("Unknown");
		}
}
