/*
 * This program is original work by Gerald Katzung <katzung@valinux.com> and
 * San Mehat <nettwerk@valinux.com>.
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999-2000, Gerald Katzung, San Mehat, and VA Linux Systems.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * VACM VA 1000 Module
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <libloose.h>
#include <kernel_module/va1000_ioctl.h>

#define NEXXUS_API_VERSION_MAJOR  1 /* move to vacm_modules.h */
#define NEXXUS_API_VERSION_MINOR  1 /* move to vacm_modules.h */

#define NO_ERROR							0
#define MALFORMATTED_MESSAGE	-110
#define UNSUPPORTED_MESSAGE		-112
#define MODULE_NOT_AVAILABLE	-120
#define NODE_NOT_FOUND				-122
#define NODE_NOT_CONFIGURED		-124
#define NODE_NOT_RESPONDING		-126

typedef __uint32_t	JOB;


static void	config_cb(__uint32_t job_id, int client_fd, char *msg);
static void	deinit_cb(__uint32_t job_id, int client_fd, char *msg);
static void	discon_cb(__uint32_t job_id, int client_fd, char *msg);
static void	ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg);

#define VA1K_ALLOW_RAW_TRANSACTIONS	0	/* issue any CMBus transaction */
#define VA1K_ALLOW_GLOBAL_WHACKING	0	/* alter globals from client */

#define TRUE	1
#define FALSE	!TRUE
#define KNOWN_UNUSED_PARAM(x) { int __bleh; __bleh = (int) x;}

#define VA1K_POWER_OFF			0			/* do_transaction() selectors */
#define VA1K_POWER_ON				1
#define VA1K_POWER_CYCLE		2
#define VA1K_HARD_RESET			3
#define VA1K_IDENTIFY				4
#define VA1K_SELECT					5
#define VA1K_MACHINE_STATUS	6
#define VA1K_EEPROM_READ		7
#define VA1K_EEPROM_WRITE		8
#define VA1K_MAX_CMD				VA1K_EEPROM_WRITE



#define VA1K_NODE_ID_SIZE  32      /* lengths of the various strings */
#define VA1K_CMB_SIZE      8 

typedef struct	nodelist
	{
	struct nodelist	*next;
	struct nodelist	*prev;
	unsigned char		identifying;
	unsigned char		selected;     /* set when selected as "console" */
	unsigned char		debug;        /* set to produce debug output for node */
	unsigned char		pad;
	int							address;
	char						node_id[VA1K_NODE_ID_SIZE];		/* node name */
	char						cmbus_address[VA1K_CMB_SIZE];	/* CM Bus addr string */
	} NODELIST;

#define VA1K_UEP_BUF_SIZE	256



extern int			errno;

static NODELIST	*nodelist_head = NULL;
static NODELIST	*nodelist_tail = NULL;

static int			g_debug_level	= 1;	/* 1=normal, 3=very verbose */
static int			g_local_node	= -1;	/* CM bus addr of nexxus node */
static int			g_smb_fd			= -1;	/* fd of piix4_smbus device */
																		/* -1: not open yet, 0: hw doesn't exist */
static int			g_table[]	= { 0x0A8D46C4, 0x0A8D45C3, 0x0A8D4FCE, 0x0A8D47C5,
															0x0D9149C7, 0x048348C6, 0x0A8D4FCE, 0x028662C8,
															0x038763C9 };
static char			g_cmd_ctr			= 0;
static int			g_timeout			= 5;	/* transaction timeout, in seconds */
static char			*g_uep_buf		= NULL;
static int			g_uep_offset	= 0;



static NODELIST	*node_create(char *node_id);
static NODELIST	*lookup_node(char *node_id);
static void			node_destroy(NODELIST	*node);
static void 		node_remove_client(NODELIST *node, int client_fd);
/*atic int			node_add_client(NODELIST *node, int client_fd);*/

static int			do_chassis_status(NODELIST *node, JOB j, int fd);
static int			do_configuration(NODELIST *node);
static int			do_debug(NODELIST *node, JOB j, int fd);
static int			do_eeprom_read(NODELIST *node, JOB j, int fd);
static int			do_eeprom_write(NODELIST *node);
static int			do_identify(NODELIST *node);
static int			do_inquiry(NODELIST *node, JOB j, int fd);
static int			do_power_ctrl(NODELIST *node, int cmd_code);
static int			do_raw(NODELIST *node, JOB j, int fd);
static int			do_select(NODELIST *node);
static int			do_transaction(int addr, int cmd_code, char *stuff);
static void			log_printf(int level, char *fmt, ...);
static int			module_init_timer_task(void *arg);
static int			piix4_transaction(smbus_pkt *p, unsigned char response,\
																	char *stuff);
static void			pkt_log_printf(void *p);
static void			save_unexpected_packet(unsigned char *bytes);
static void			send_final_response(int err, JOB job_id, int reply_fd);
static void			setup_smbus_pkt(smbus_pkt *p, int addr, int cmd, char *stuff,\
																unsigned char *response);
static int			timeout(time_t timeout);
static int			va1k_module_init(void);



int main(int argc, char *argv[])
{
	KNOWN_UNUSED_PARAM(argc);
	KNOWN_UNUSED_PARAM(argv);

	lm_init();
	lm_nexxus_connect();
	if (lm_register("VA1000",
									"VA 1000 Module",
									"Module to support VA Linux 1000 clusters.",
									"Jerry Katzung (katzung@valinux.com)",
									"VA1000",
									NEXXUS_API_VERSION_MAJOR,
									NEXXUS_API_VERSION_MINOR) < 0)
		{
		printf("Unable to register module with Nexxus\n");
		exit(-1);
		}

	if (va1k_module_init() < 0)
		lm_timer_add(1, module_init_timer_task, NULL);	/* keep trying... */

	lm_main_loop(config_cb, deinit_cb, discon_cb, ipc_cb);
	exit(0);
}

static void config_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;
	char			*cmd_module;
	char			*node_id;

	log_printf(2, "config_cb(job_id=%d, client_fd=%d, msg=%s)",
							job_id, client_fd, msg);

	cmd_module = strtok(msg, ":");
	node_id = strtok(NULL, ":");

	node = lookup_node(node_id);

	if (!strcasecmp(cmd_module, "NODE"))
		{
		if (node != NULL)
			log_printf(1, "Node '%s' already exists!", node_id);
		else
			node = node_create(node_id);
		}
	else if (!strcasecmp(cmd_module, "END_NODE"))
		{
		}
	else if (!strcasecmp(cmd_module, "ADDITION"))
		{
		if (node != NULL)
			log_printf(1, "Node '%s' already exists!", node_id);
		else
			node = node_create(node_id);
		}
	else if (!strcasecmp(cmd_module, "DELETION"))
		{
		node_destroy(node);
		}
	else if (!strcasecmp(cmd_module, "RENAME"))
		{
		char	*new_name;
		
		if (!(new_name = strtok(NULL,":")))
			{
			log_printf(1, "[VA1000] Malformatted RENAME config message.");
			return;
			}
		strncpy(node->node_id, new_name, (VA1K_NODE_ID_SIZE-1));
		}
	else if (!strcasecmp(cmd_module, "GLOBAL"))
		{
		/* There are no globals we care about. */	
		}
	else if (!strcasecmp(cmd_module, "VA1000"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			log_printf(1, "[VA1000] Malformatted VA1000 variable config message.");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			log_printf(1, "[VA1000] Malformatted VA1000 value config message.");
			return;
			}
		/* Process any local variables here */
		if (!strcasecmp(variable, "CMBUS_ADDRESS"))
			{
			strncpy(node->cmbus_address, value, (VA1K_CMB_SIZE-1));
			node->address = atoi(node->cmbus_address);
			if (node->address == 0)
				node->address = -1;
			}
		else
			log_printf(1, "[VA1000] Unknown local module variable '%s'.", variable);

		}
	else
		log_printf(1, "[VA1000] Unknown config event <%s>.", cmd_module);
}

static void deinit_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);

	lm_nexxus_disconnect();
	exit(0);
}

static void discon_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;

	log_printf(2, "discon_cb(job_id=%d, client_fd=%d, msg=%s)",
							job_id, client_fd, msg);

	node = nodelist_head;
	while(node)
		{
		node_remove_client(node, client_fd);
		node = node->next;
		}
}

static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;
	char			*cmd;
	char			*scratch;
	int				error		= NO_ERROR;					/* assume all is fine */
	
	log_printf(2, "ipc_cb(node_id=%s, job_id=%d, client_fd=%d, msg=%s)",
							node_id, job_id, client_fd, msg);

	if (!(cmd = strtok(msg, ":")) ||			/* grab command string and */
			!(scratch = strtok(NULL, ":")))		/* consume globbable node name */
		{
		send_final_response(MALFORMATTED_MESSAGE, job_id, client_fd);
		return;
		}

	/*
		The value of g_smb_fd can mean a number of different things:
		g_smb_fd > 0: SMBus driver has been opened successfully.
		g_smb_fd = 0: SMBus device does NOT exist - things will never work.
		g_smb_fd =-1: SMBus device in use - things should work when hw frees up.
		In all the error cases, we report that the module is not available.

		It is also possible that the SMBus interrupts are not getting through,
		which would have made it impossible to determine the local node's
		CMBus address.  The low-level transaction will return the appropriate
		error in that case.
	*/

	if (g_smb_fd <= 0)
		{
		send_final_response(MODULE_NOT_AVAILABLE, job_id, client_fd);
		return;
		}

	node = lookup_node(node_id);	/* wouldn't be here if node_id was invalid */

	/* Except for "configuration" messages, the node's address must be known */

	if ((node->address == -1) && (strcasecmp(cmd, "CONFIGURATION") != 0))
		{
		send_final_response(NODE_NOT_CONFIGURED, job_id, client_fd);
		return;
		}

	if			(!strcasecmp(cmd, "CONFIGURATION"))
		{
		error = do_configuration(node);
		}
	else if (!strcasecmp(cmd, "INQUIRY"))
		{
		error = do_inquiry(node, job_id, client_fd);
		}
	else if (!strcasecmp(cmd, "POWER_OFF"))
		{
		error = do_power_ctrl(node, VA1K_POWER_OFF);
		}
	else if (!strcasecmp(cmd, "POWER_ON"))
		{
		error = do_power_ctrl(node, VA1K_POWER_ON);
		}
	else if (!strcasecmp(cmd, "POWER_CYCLE"))
		{
		error = do_power_ctrl(node, VA1K_POWER_CYCLE);
		}
	else if (!strcasecmp(cmd, "HARD_RESET"))
		{
		error = do_power_ctrl(node, VA1K_HARD_RESET);
		}
	else if (!strcasecmp(cmd, "CHASSIS_STATUS"))
		{
		error = do_chassis_status(node, job_id, client_fd);
		}
	else if (!strcasecmp(cmd, "IDENTIFY"))
		{
		error = do_identify(node);
		}
	else if (!strcasecmp(cmd, "SELECT"))
		{
		error = do_select(node);
		}
	else if (!strcasecmp(cmd, "EEPROM_READ"))
		{
		error = do_eeprom_read(node, job_id, client_fd);
		}
	else if (!strcasecmp(cmd, "EEPROM_WRITE"))
		{
		error = do_eeprom_write(node);
		}
	else if (!strcasecmp(cmd, "DEBUG"))
		{
		error = do_debug(node, job_id, client_fd);
		}
	else if (!strcasecmp(cmd, "RAW"))
		{
		error = do_raw(node, job_id, client_fd);
		}
	else
		{
		error = UNSUPPORTED_MESSAGE;
		}

	send_final_response(error, job_id, client_fd);
	return;
}

static NODELIST	*node_create(char *node_id)
{
	NODELIST	*new;

	if (!(new = (NODELIST	*) malloc(sizeof(NODELIST))))
		return(0);
	memset(new, 0, sizeof(NODELIST));
	new->address = -1;								/* initialized to an invalid value */
	if (!nodelist_head)
		nodelist_head = nodelist_tail = new;
	else
		{
		nodelist_tail->next = new;
		new->prev = nodelist_tail;
		nodelist_tail = new;
		}
	strcpy(new->node_id, node_id);
	return(new);
}

static NODELIST	*lookup_node(char *node_id)
{
	NODELIST	*scan;

	if (!node_id)
		return(NULL);
	scan = nodelist_head;
	while(scan)
		{
		if (!strcmp(node_id, scan->node_id))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

static void	node_destroy(NODELIST	*node)
{
  if (node == nodelist_head)
    {
    nodelist_head = nodelist_head->next;
    if (nodelist_head)
      nodelist_head->prev = NULL;
    else
      nodelist_tail = NULL;
    } 
  else if (node == nodelist_tail)
    {
    nodelist_tail = nodelist_tail->prev;
    if (nodelist_tail)
      nodelist_tail->next = NULL;
    else
      nodelist_head = NULL;
    } 
  else
    {
    node->next->prev = node->prev;
    node->prev->next = node->next;
    }
  free(node);
}


#if 0
static int node_add_client(NODELIST *node, int client_fd)
{
	KNOWN_UNUSED_PARAM(node);
	KNOWN_UNUSED_PARAM(client_fd);

	return NO_ERROR;
}
#endif


static void node_remove_client(NODELIST *node, int client_fd)
{
	KNOWN_UNUSED_PARAM(node);
	KNOWN_UNUSED_PARAM(client_fd);
}


/*
	This routine attempts to open up the SMBus hardware device, and determine
	the CMBus address of the node that it is running on.  If all goes well,
	g_smb_fd and g_local_node are set to positive values.  If the device
	doesn't exist (wrong hardware) then g_smb_fd is set to zero, indicating
	that it can never be opened successfully.  If the SMBus device exists, but
	cannot be opened (because it is in use) g_smb_fd remains at -1.  In this
	case, a timer will be set up to keep trying to open the driver.
*/

int va1k_module_init(void)
{
  char  device[32] = "/dev/va1000_smbus";
	char	stuff[4] = { 0,0,0,0 };
	char	ch, cl;
	int		err;

  log_printf(3, "Initialization commencing");
  
	if (g_uep_buf == NULL)
		if ((g_uep_buf = malloc(VA1K_UEP_BUF_SIZE)))
			memset(g_uep_buf, 0, VA1K_UEP_BUF_SIZE);
	
  g_smb_fd = open(device, O_RDONLY);
  if (g_smb_fd < 0) {
		if (errno == EBUSY)
    	log_printf(1, "Failed to open %s: device already in use!", device);
		else if (errno == ENOENT) {
			log_printf(1, "Failed to open %s: device does not exist!", device);
			g_smb_fd = 0;								/* remember that hw doesn't exist */
		} else {
    	log_printf(1, "Failed to open %s. (%s)", device, strerror(errno));
			g_smb_fd = 0;	/* probably won't pay to keep trying to open */
		}

    return g_smb_fd;
  } 
 
	stuff[0] = 2;
	stuff[1] = 0;
	stuff[2] = 2;
	err = do_transaction(0, VA1K_EEPROM_READ, stuff);
	if (err || (stuff[0] != 1))
		err = -1;
	else {
		ch = stuff[1];
		stuff[0] = 2;
		stuff[1] = 0;
		stuff[2] = 3;
		err = do_transaction(0, VA1K_EEPROM_READ, stuff);
		if (err || (stuff[0] != 1))
			err = -1;
		else
			cl = stuff[1];
	}

	if (err) {
		log_printf(1, "Failed to get local node's address!");
	} else {
		g_local_node = (ch << 8) | cl;
	}

	log_printf(1, "Local node address is %d", g_local_node);
  log_printf(3, "Initialization complete");
  return NO_ERROR;
} 


/* This VACM module requires exclusive access to the SMBus controller in
   the PIIX4.  If unable to open it, just keep trying.    */

int module_init_timer_task(void *arg)
{
	KNOWN_UNUSED_PARAM(arg);

	if (g_smb_fd > 0)
		return FALSE;					/* if we've opened the driver, we're in business */

	if (g_smb_fd == 0)
		return FALSE;					/* nonexistant hardware - this will never work */

	if (va1k_module_init() < 0)
		return TRUE;					/* may just be busy hardware - keep trying */
	else
		return FALSE;					/* anything else - give up */
}


void log_printf(int level, char *fmt, ...)
{
  char  str[256];
  va_list ap;
  
  if (g_debug_level >= level) {
    va_start(ap, fmt); 
    vsnprintf(str, 256, fmt, ap);
    va_end(ap);
		lm_log(str);
  } 
} 

void pkt_log_printf(void *ppp)
{
  smbus_pkt *p = (smbus_pkt *) ppp;
  char      str[80];
  
  snprintf(str, 80,
					"%02X %02X %02X %02X %02X (%02X %02X %02X %02X %02X %02X %02X %02X)",
              p->status, p->type, p->len, p->addr, p->cmd,
              p->data[0], p->data[1], p->data[2], p->data[3],
              p->data[4], p->data[5], p->data[6], p->data[7]);
  log_printf(3, "%s", str);
} 

void send_final_response(int err, JOB j, int fd)
{
	char	rsvp[128];

	switch (err) {

	case NO_ERROR:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_COMPLETED", j, fd);
		break;

	case MALFORMATTED_MESSAGE:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE", j, fd);
		break;

	case UNSUPPORTED_MESSAGE:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:UNSUPPORTED_MESSAGE", j, fd);
		break;

	case MODULE_NOT_AVAILABLE:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:MODULE_NOT_AVAILABLE", j, fd);
		break;

	case NODE_NOT_CONFIGURED:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:NODE_NOT_CONFIGURED", j, fd);
		break;

	case NODE_NOT_FOUND:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:NODE_NOT_FOUND", j, fd);
		break;

	case NODE_NOT_RESPONDING:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:NODE_NOT_RESPONDING", j, fd);
		break;

	default:
		snprintf(rsvp, 128, "%d:%d:FOR:JOB_ERROR:%d", j, fd, err);
		log_printf(1, "Unexpected error response: %d", err);

	}

  lm_send_to_nexxus(rsvp);
}


int do_configuration(NODELIST *node)
{
	char	msg[80];
	char	*str;
	int		addr;

	if (!(str = strtok(NULL, ":")))		/* get the CMBus address string */
		return MALFORMATTED_MESSAGE;

	addr = atoi(str);
	if (addr <= 0) {
		log_printf(1, "[VA1000] Invalid CMBus address: %s", str);
		return MALFORMATTED_MESSAGE;
	}

	strncpy(node->cmbus_address, str, (VA1K_CMB_SIZE-1));
	node->address = addr;

	snprintf(msg, 80, "0:0:SET:%s:VA1000:CMBUS_ADDRESS:%s", node->node_id, str);
	lm_send_to_nexxus(msg);

	return NO_ERROR;
}


int do_inquiry(NODELIST *node, JOB j, int fd)
{
	char	rsvp[128];

	snprintf(rsvp, 128, "%d:%d:FOR:INQUIRY:%s:%s",
					j, fd, node->node_id, node->cmbus_address );

	lm_send_to_nexxus(rsvp);

	return NO_ERROR;
}


int do_chassis_status(NODELIST *node, JOB j, int fd)
{
	char	rsvp[128];
	int		err, i;
	char	*v[8];
	char	yes[] = "YES";
	char	no[]	= "NO";
	char	stuff[] = { 0,0,0,0 };

	if ((err = do_transaction(node->address, VA1K_MACHINE_STATUS, stuff)))
		return err;

	log_printf(3, "chs_sts: stuff[0..3] = 0x%02X, 0x%02X, 0x%02X, 0x%02X",
							stuff[0], stuff[1], stuff[2], stuff[3]);

	for (i = 0 ; i < 8 ; i++)
		v[i] = no;

	if (stuff[0] == 0)
		stuff[1] = 0x01;	/* it's there, but didn't return data - assume it's on */

	if ((stuff[1] & 0x01) == 0)	/* no power: make sure temp_alert is clear */
		stuff[1] &= ~0x10;

	for (i = 0 ; i < 8 ; i++)
		if (stuff[1] & (1 << i))
			v[i] = yes;

	snprintf(rsvp, 128,
					"%d:%d:FOR:CSTATUS:%s:%s:%s:%s:%s:%s:%s:%s",
					j, fd, v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7] );

	lm_send_to_nexxus(rsvp);

	return NO_ERROR;
}


int do_power_ctrl(NODELIST *node, int cmd_code)
{
  int		err, err2;
  
  switch (cmd_code) {

  case VA1K_POWER_OFF:
  case VA1K_POWER_ON:
  case VA1K_HARD_RESET:
    err = do_transaction(node->address, cmd_code, NULL);
    break;

  case VA1K_POWER_CYCLE:
    err  = do_transaction(node->address, VA1K_POWER_OFF, NULL);
    err2 = do_transaction(node->address, VA1K_POWER_ON, NULL);
    err = (err ? err : err2);
		break;

	default:
		log_printf(1, "Unexpected power_ctrl selector");
		err = UNSUPPORTED_MESSAGE;	/* if we got here, ipc_cb made a mistake */
  } 
  
	return err;
} 


int do_identify(NODELIST *node)
{ 
  char    *option;
  int     cmd_code	= VA1K_IDENTIFY;
  int     err				= NO_ERROR;
  
  log_printf(3, ">>> do_identify");
  
  if (!(option = strtok(NULL, ":")))
		return MALFORMATTED_MESSAGE;
  
  if (!(strncasecmp(option, "OFF", 3))) {
    node->identifying = FALSE;
    //node->selected = FALSE;
    cmd_code = VA1K_SELECT;
  } else if (!(strncasecmp(option, "ON", 2))) {
    node->identifying = TRUE;
    //node->selected = FALSE;
    cmd_code = VA1K_IDENTIFY;
  } else
    err = MALFORMATTED_MESSAGE;
  
	if (err == NO_ERROR)
  	err = do_transaction(node->address, cmd_code, NULL);
  
  log_printf(3, "<<< do_identify");

	return err;
}


int do_select(NODELIST *node)
{ 
  int     err		= NO_ERROR;
	int			err2	= NO_ERROR;
  
  log_printf(3, ">>> do_select");

	/* An identifying node needs to be selected twice - once to stop
		 identifying, and then once to be selected. */

	if (node->identifying) {
		node->identifying = FALSE;
		err = do_transaction(node->address, VA1K_SELECT, NULL);
	}

	err2 = do_transaction(node->address, VA1K_SELECT, NULL);

	err = (err ? err : err2);

  log_printf(3, "<<< do_select");

	return err;
}


int do_eeprom_read(NODELIST *node, JOB j, int fd)
{
	char		*addr_str;
	int			addr;
	int			err = NO_ERROR;
	char		args[4] = { 0,0,0,0 };
	char		rsvp[64];

	if (!(addr_str = strtok(NULL, ":")))
		return MALFORMATTED_MESSAGE;

	addr = atoi(addr_str);
	if (addr <= 0) {
		log_printf(1, "[VA1000] Invalid EEPROM address: %s", addr_str);
		return MALFORMATTED_MESSAGE;
	}

	args[0] = 2;
	args[1] = addr >> 8;
	args[2] = addr & 0xFF;

	err = do_transaction(node->address, VA1K_EEPROM_READ, args);

	if (err)
		return err;

	if (args[0] != 1) {
		log_printf(1, "[VA1000] EEPROM read returned %d bytes.", args[0]);
		return NODE_NOT_RESPONDING;
	}

	snprintf(rsvp, 64, "%d:%d:FOR:EEPROM_READ:0x%02X", j, fd, args[1]);
	lm_send_to_nexxus(rsvp);
	return err;	
}


int do_eeprom_write(NODELIST *node)
{
	char		*addr_str, *value_str;
	int			addr, value;
	int			err = NO_ERROR;
	char		args[4] = { 0,0,0,0 };

	if (!(addr_str = strtok(NULL, ":")))
		return MALFORMATTED_MESSAGE;

	if (!(value_str = strtok(NULL, ":")))
		return MALFORMATTED_MESSAGE;

	addr = atoi(addr_str);
	if (addr <= 0) {
		log_printf(1, "[VA1000] Invalid EEPROM address: %s", addr_str);
		return MALFORMATTED_MESSAGE;
	}

	value = atoi(value_str);
	if (value <= 0) {
		log_printf(1, "[VA1000] Invalid value: %s", value_str);
		return MALFORMATTED_MESSAGE;
	}

	args[0] = 3;
	args[1] = addr >> 8;
	args[2] = addr & 0xFF;
	args[3] = value & 0xFF;

	err = do_transaction(node->address, VA1K_EEPROM_WRITE, args);

	return err;	
}


/*
	On entry, the command and globbable node name have been stripped from
	the message.  What should remain includes:
		GET:<variable>
		SET:<variable>:<value> # VA1K_ALLOW_GLOBAL_WHACKING must be true
		DUMP
*/

int do_debug(NODELIST *node, JOB j, int fd)
{
	char			*subcmd		= strtok(NULL, ":");
	char			*variable	= strtok(NULL, ":");
	char			*value		= strtok(NULL, ":");
	char			varstr[]	= "%d:%d:FOR:DEBUG:%s:%d";
	char			nodestr[]	= "%d:%d:FOR:DEBUG:%s:%s:%d:%c:%c:%c";
	char			uepstr[]	= "%d:%d:FOR:DEBUG:uep:%02X:";
	char			str[80];
	char			str2[16];
	NODELIST	*n				= nodelist_head;
	int				err				= NO_ERROR;
	int				i, ii;

	KNOWN_UNUSED_PARAM(node);

	if (subcmd == NULL)
		return MALFORMATTED_MESSAGE;

	if (!strcasecmp(subcmd, "GET")) {
		log_printf(3, "DEBUG:GET");
		if (variable == NULL)
			return MALFORMATTED_MESSAGE;

	} else if (!strcasecmp(subcmd, "SET")) {
		log_printf(3, "DEBUG:SET");
		if ((variable == NULL) || (value == NULL))
			return MALFORMATTED_MESSAGE;
#if VA1K_ALLOW_GLOBAL_WHACKING

#else
			return UNSUPPORTED_MESSAGE;
#endif
	} else if (!strcasecmp(subcmd, "DUMP")) {
		log_printf(3, "DEBUG:DUMP");
		snprintf(str, 80, varstr, j, fd, "g_local_node", g_local_node);
		lm_send_to_nexxus(str);
		snprintf(str, 80, varstr, j, fd, "g_debug_level", g_debug_level);
		lm_send_to_nexxus(str);
		snprintf(str, 80, varstr, j, fd, "g_smb_fd", g_smb_fd);
		lm_send_to_nexxus(str);
		snprintf(str, 80, varstr, j, fd, "g_cmd_ctr", g_cmd_ctr);
		lm_send_to_nexxus(str);
		snprintf(str, 80, varstr, j, fd, "g_timeout", g_timeout);
		lm_send_to_nexxus(str);
		
		while (n) {
			snprintf(str, 80, nodestr, j, fd, n->node_id,
								n->cmbus_address, n->address,
								(n->identifying ? 'I' : 'i'),
								(n->selected ? 'S' : 's'),
								(n->debug ? 'D' : 'd') );
			lm_send_to_nexxus(str);
			n = n->next;
		}

		if (g_uep_buf != NULL) {
			i = 0;
			while ((i < VA1K_UEP_BUF_SIZE) && (g_uep_buf[i] != 0)) {
				snprintf(str, 80, uepstr, j, fd, g_uep_buf[i]);
				for (ii = 1 ; ii <= g_uep_buf[i] ; ii++) {
					sprintf(str2, "%02X ", g_uep_buf[i+ii]);
					strcat(str, str2);
				}
				lm_send_to_nexxus(str);
				i += (g_uep_buf[i] + 1);
			}
		}

	} else {
		log_printf(1, "Unsupported debug subcommand: %s", subcmd);
		err = UNSUPPORTED_MESSAGE;
	}
	return err;
}


/*
	Perform a transaction over the SMBus.  This routine builds up and sends
	out the request, and then gathers the response packet.

	Any additional bytes are provided in "stuff", and stuff[0] contains a
	count of additional bytes.  If there are no additional bytes, stuff may
	be NULL.  If stuff is provided, additional response bytes are stored
	there as well, with stuff[0] set appropriately.  The array must be at
	least 4 bytes long.
*/

int do_transaction(int addr, int cmd_code, char *stuff)
{
  smbus_pkt				pkt, *p = &pkt;
	unsigned char		response;
  
	/* If the local node's CMBus address has not been determined, and we
		 are not sending a command to the local node (addr==0), then we fail
		 the request to the client, and send a note to the log.
		 This usually happens if SMBus interrupts are not getting through. */

	if (addr && (g_local_node == -1)) {
		log_printf(1, "Local node address is unknown (-1)");
		return MODULE_NOT_AVAILABLE;
	}
  
	setup_smbus_pkt(p, addr, cmd_code, stuff, &response);

	return piix4_transaction(p, response, stuff);
}


/*
	Perform a transaction over the SMBus.  This routine sends out the request,
	and then gathers the response packet.

	Any additional bytes are returned in "stuff", if provided, and stuff[0]
	contains a count of additional bytes.  The array must be at least 4 bytes
	long.
*/

int piix4_transaction(smbus_pkt *p, unsigned char response, char *stuff)
{
  int             err;
  int             state;
	int							i;
	unsigned char		this_cmd;
	unsigned char		this_dst_hi;
	unsigned char		this_dst_lo;
  unsigned char   expected;
	unsigned char		bytes[32];
	time_t					end_time;

	this_cmd = p->cmd;									/* remember the transaction tag */
	this_dst_hi = p->data[0];						/* and target destination addr */
	this_dst_lo = p->data[1];

	if (stuff)
		stuff[0] = 0;

  if ((err = ioctl(g_smb_fd, VA1000_SMBUS_HOST_TRANSFER, p)) != NO_ERROR)
    return err;

	end_time = time(NULL) + g_timeout;				/* calculate timeout */
    
  log_printf(3, "Waiting to retire host transaction...");
  state = 1;
  while ((state == 1) && !timeout(end_time)) {
    err = ioctl(g_smb_fd, VA1000_SMBUS_RETIRE, p);
    if (err == NO_ERROR) {
      if (p->type != 0xFF) {
				if (p->cmd == this_cmd) {
        	state = 2;        /* host transaction retired */
					if (g_debug_level >= 3)
						pkt_log_printf(p);
				} else {
					log_printf(1, "Unexpected (different) host transaction!");
					pkt_log_printf(p);
				}
      } else {
        log_printf(1, "Unexpected slave transaction!");
				pkt_log_printf(p);
			}
    }
  } 
  
  log_printf(3, "Waiting to retire slave transactions...");

	expected = 31;														/* not likely */
  while ((state == 2) && !timeout(end_time)) {
    err = ioctl(g_smb_fd, VA1000_SMBUS_RETIRE, p);
    if (err == NO_ERROR) {
      if (p->type == 0xFF) {
				if (p->data[0] & 0x40) {
					memset(bytes, 0, 32);
					expected = p->data[0] & 0x1F;
					bytes[0] = 1;
					bytes[1] = p->data[1];
				} else {
					bytes[p->data[0] & 0x1F] = p->data[1];
					bytes[0]++;
					if (bytes[0] == expected)
          	state = 3;                      /* last expected pkt retired */
				}
				if (g_debug_level >= 3)
					pkt_log_printf(p);
      } else {
        log_printf(1, "Unexpected host transaction!");
      	pkt_log_printf(p); 
      } 
    }
		if (state == 3) {
			if ((bytes[3] != this_dst_hi) || (bytes[4] != this_dst_lo)
					|| (bytes[5] != response)) {
				if (bytes[5] == 0x8A)
					return NODE_NOT_RESPONDING;
				else {
					log_printf(1, "Unexpected response packet!");
					save_unexpected_packet(bytes);
					memset(bytes, 0, 32);
					state = 2;
				}
			}
 		}
  } 

	if (state == 3) {
		if (stuff && (bytes[0] > 5)) {
			if (bytes[0] > 8)
				log_printf(1, "Unexpectedly large packet: %d bytes", bytes[0]);
			else {
				stuff[0] = bytes[0] - 5;
				for (i = 0 ; i < stuff[0] ; i++)
					stuff[1+i] = bytes[6+i];
			}
		}
  	return NO_ERROR;
	} else {
		log_printf(1, "Timed out during transaction!");
		return MODULE_NOT_AVAILABLE;
	}
}


void save_unexpected_packet(unsigned char *bytes)
{
	char	msg[128];
	char	str[16];
	int		i;

	snprintf(msg, 128, "Unexpected packet: len=%d, contains { ", bytes[0]);
	
	for (i = 1 ; i <= bytes[0] ; i++) {
		sprintf(str, "%02X ", (unsigned char) bytes[i]);
		strcat(msg, str);
	}
	sprintf(str, "}");
	strcat(msg, str);

	log_printf(1, msg);

	if (g_uep_buf == NULL)
		return;

	if (g_uep_offset + bytes[0] + 1 < VA1K_UEP_BUF_SIZE) {
		memcpy(&g_uep_buf[g_uep_offset], bytes, bytes[0] + 1);
		g_uep_offset += bytes[0] + 1;
	} else {
		snprintf(msg, 128, "Unexpected packet buffer full!");		
		log_printf(1, msg);
	}
}


int timeout(time_t timeout)
{
	if (time(NULL) <= timeout)
		return FALSE;
	else
		return TRUE;
}






void setup_smbus_pkt(smbus_pkt *p, int addr, int cmd_code, char *stuff,
											unsigned char *response)
{
  int	val		= 0;
	int offset = 24;
	int	i;
 
	if (addr && (addr != g_local_node)) {
		val = addr;
		offset -= 16;
	}
  
  memset(p, 0, sizeof(smbus_pkt));
  p->size = sizeof(smbus_pkt);
  p->type = PIIX4_BLOCK_WRITE;
  p->len  = 5;
  p->addr = 0xE6;
  p->cmd  = ++g_cmd_ctr;
  p->data[0] = val / 256;
  p->data[1] = val % 256;
 
	if ((cmd_code < 0) || (cmd_code > VA1K_MAX_CMD)) {
    log_printf(1, "Unsupported cmd_code: %d", cmd_code);
		cmd_code = VA1K_MACHINE_STATUS;
	}

	p->data[4] = (g_table[cmd_code] >> offset) & 0xFF;
	*response = (g_table[cmd_code] >> (offset - 8)) & 0xFF;
	
	if (stuff && (stuff[0] >= 1) && (stuff[0] <= 3))
		for (i = 1 ; i <= stuff[0] ; i++) {
			p->data[4+i] = stuff[i];
			p->len++;
		}

	log_printf(3, "setup_smbus_pkt: cmd_code = %d, offset = %d, cmd = 0x%02X, response = 0x%02X", cmd_code, offset, p->data[4], *response);
}


int do_raw(NODELIST *node, JOB j, int fd)
{
#if VA1K_ALLOW_RAW_TRANSACTIONS

	int							err;
	int							addr;
	int							i;
	unsigned char		cmd;
	unsigned char		response;
	char						args[4] = { 0,0,0,0 };
	char						*msg;
	smbus_pkt				pkt, *p = &pkt;
	unsigned int		v[8] = { 0,0,0,0,0,0,0,0 };
	char						rsvp[80], nstr[16];

	KNOWN_UNUSED_PARAM(node);
	KNOWN_UNUSED_PARAM(j);
	KNOWN_UNUSED_PARAM(fd);

	if (!(msg = strtok(NULL, ":")))
		return MALFORMATTED_MESSAGE;

	log_printf(2, "do_raw(msg=%s)", msg);

	i = sscanf(msg, "%x %x %x %x %x %x %x",
							&v[0], &v[1], &v[2], &v[3], &v[4], &v[5], &v[6]);

	if (i < 3)
		return MALFORMATTED_MESSAGE;

	addr			= v[0] & 0xFFFF;
	cmd				= v[1] % 256;
	response	= v[2] % 256;
	args[0]		= v[3] % 256;
	args[1]		= v[4] % 256;
	args[2]		= v[5] % 256;
	args[3]		= v[6] % 256;

	log_printf(3, "do_raw: addr = %x, cmd = %x, response = %x",
							addr, cmd, response);
	log_printf(3, "do_raw: args = { %x, %x, %x, %x }",
							args[0], args[1], args[2], args[3]);

	memset(p, 0, sizeof(smbus_pkt));
  p->size = sizeof(smbus_pkt);
  p->type = PIIX4_BLOCK_WRITE;
  p->len  = 5;
  p->addr = 0xE6;
  p->cmd  = ++g_cmd_ctr;
	p->data[0] = addr / 256;
	p->data[1] = addr % 256;
	p->data[4] = cmd;

	if ((args[0] >= 1) && (args[0] <= 3))
		for (i = 1 ; i <= args[0] ; i++) {
			p->data[4+i] = args[i];
			p->len++;
		}

	err = piix4_transaction(p, response, args);
	
	if ((args[0] >= 1) && (args[0] <= 3)) {
		snprintf(rsvp, 80, "%d:%d:FOR:RAW", j, fd);
		for (i = 1 ; i <= args[0] ; i++) {
			snprintf(nstr, 16, "%c0x%02X", ((i==1) ? ':' : ' '), args[i]);
			strcat(rsvp, nstr);
		}
		lm_send_to_nexxus(rsvp);
	}

	return err;

#else

	KNOWN_UNUSED_PARAM(node);
	KNOWN_UNUSED_PARAM(j);
	KNOWN_UNUSED_PARAM(fd);

	return UNSUPPORTED_MESSAGE;

#endif
}
